# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2001-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, subprocess, os, fcntl
from io import BytesIO

import PyQt5.QtCore as qtcore

import editobj3
import songwrite3, songwrite3.globdef as globdef

_p            = None
_midi         = ""
_loop         = 0
_tracker      = None
_last_tracker = None
_output       = b""

def is_playing(): return _p and (_p.poll() is None)

def stop():
  global _p, _tracker
  if not _p is None:
    if _tracker:
      _tracker(-1)
      _tracker = None
      
    try:
      _p.terminate()
      _p.kill()
    except OSError:
      sys.excepthook(*sys.exc_info())
    _p  = None
    
def play(midi, loop = 0, tracker = None):
  global _p, _midi, _loop, _tracker, _last_tracker, _output
  
  stop()
  
  if not globdef.config.DISPLAY_PLAY_BAR: tracker = None
  _midi    = midi
  _loop    = loop
  _tracker = _last_tracker = tracker
  _output  = b""
  
  _play()
  
def _play():
  global _p, _midi, _loop, _tracker
  if globdef.config.MIDI_USE_TEMP_FILE:
    import tempfile
    file = tempfile.mktemp(".midi")
    open(file, "w").write(_midi)
    _p = p = subprocess.Popen(globdef.config.MIDI_COMMAND % file, shell = True, stdin = subprocess.PIPE, stdout = subprocess.PIPE)
  else:
    _p = p = subprocess.Popen(globdef.config.MIDI_COMMAND, shell = True, stdin = subprocess.PIPE, stdout = subprocess.PIPE)
    _p.stdin.write(_midi)
    _p.stdin.close()
  fcntl.fcntl(p.stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)
  
  
  try: _play.qt_timer.stop()
  except: pass
  _play.qt_timer = qtcore.QTimer()
  _play.qt_timer.timeout.connect(on_timer)
  _play.qt_timer.setSingleShot(False)
  if _tracker: _play.qt_timer.setInterval( 50)
  else:        _play.qt_timer.setInterval(300)
  _play.qt_timer.start()
  

def on_timer():
  global _tracker, _output
  
  try:
    if _p is None:
      _play.qt_timer.stop() #return 0 # Stop was called
      return
    
    r = _p.poll()
    
    if r is None:
      if _tracker:
        output = _p.stdout.read()
        
        if output:
          lines   = (_output + output).split(b"\n")
          _output = lines[-1]
          if len(lines) > 1:
            line = lines[-2]
            try:
              if b":" in line: t = int(line[line.rfind(b":") + 3 :])
              else:            t = int(line)
            except: return # 1
            _tracker(t)
            if t == -1:
              _tracker = None # Tracking has reached its end !
              stop()
              if _loop:
                _tracker = _last_tracker
                _play()
                
    else:
      _play.qt_timer.stop()
      
      if (r == 0) and _loop:
        _tracker = _last_tracker
        _play()
        
  except: sys.excepthook(*sys.exc_info())


noteplayer = None

def play_note(instrument, value):
  if not globdef.config.PLAY_AS_TYPING: return
  
  stop()
  
  global noteplayer
  if not noteplayer:
    noteplayer = songwrite3.model.Song()
    noteplayer.partitions.append(songwrite3.model.Partition(noteplayer))
    noteplayer.partitions[0].notes.append(songwrite3.model.Note(noteplayer.partitions[0], 0, 48, 0))
  noteplayer.partitions[0].instrument = instrument
  noteplayer.partitions[0].notes[0].value = value
  
  play(songwrite3.midi.song_2_midi(noteplayer))



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2001-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, time, getpass, os, os.path, locale, unicodedata, subprocess
from io import StringIO

import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtGui     as qtgui

import editobj3, editobj3.undoredo as undoredo, editobj3.introsp as introsp, editobj3.field as field, editobj3.editor as editor, editobj3.editor_qt as editor_qt
from editobj3.observe import *

import songwrite3, songwrite3.globdef as globdef, songwrite3.model as model, songwrite3.player as player
import songwrite3.__editobj3__
import songwrite3.midi

editor_qt.MENU_LABEL_2_IMAGE.update({
  "Add..."                             : "list-add",
  "Remove"                             : "list-remove",
  "Move up"                            : "go-up",
  "Move down"                          : "go-down",
  "Play from the beginning"            : "media-playback-start",
  "Play from the beginning in loop"    : "media-playback-start",
  "Play from here"                     : "media-playback-start",
  "Play from here in loop"             : "media-playback-start",
  "Stop playing"                       : "media-playback-stop",
  "Edit..."                            : "document-properties",
  "Songbook properties..."             : "document-properties",
  "Song and instruments properties..." : "document-properties",
  "Instrument properties..."           : "document-properties",
  "Note properties..."                 : "document-properties",
  "Save songbook as..."                : "document-save-as",
  "preview_print"                      : "document-print",
  "preview_print_songbook"             : "document-print",
  "New window..."                      : "window-new",
  "New song"                           : "document-new",
  "New songbook"                       : "document-new",
  "Select all"                         : "edit-select-all",
  "Paste last selection"               : "edit-paste",
  "About..."                           : "help-about",
  "Manual..."                          : "help-contents",
})

from songwrite3.canvas import Canvas

app      = qtwidgets.QApplication(sys.argv)
mainloop = app.exec
app.setWindowIcon(qtgui.QIcon(os.path.join(globdef.DATADIR, "songwrite3_64x64.png")))

def remove_accents(s): return unicodedata.normalize("NFKD", s).encode("ascii", "ignore").decode("ascii")


APPS = {}

class App(editor_qt.QtBaseDialog):
  def __init__(self, song = None):
    self.init_window(False)
    
    self.undo_stack       = undoredo.Stack(globdef.config.NUMBER_OF_UNDO)
    self.updating_menu    = False
    self.partition_editor = None
    self.mesure_editor    = None
    self.note_editor      = None
    self.filename         = None
    
    self.build_default_menubar()
    
    #file_menu = self.add_to_menu(self.menubar, 1, "File", self.on_file_menu)
    file_menu = self.file_menu
    self.add_to_menu(file_menu, 0, "New window..."      , self.on_new_app     , pos = 0)
    self.add_to_menu(file_menu, 0, "New song"           , self.on_new         , pos = 1, accel = "C-N")
    self.add_to_menu(file_menu, 0, "New songbook"       , self.on_new_songbook, pos = 2)
    self.add_to_menu(file_menu, 0, "Open..."            , self.on_open        , pos = 3, accel = "C-O")
    self.add_to_menu(file_menu, 0, "Save"               , self.on_save        , pos = 4, accel = "C-S")
    self.add_to_menu(file_menu, 0, "Save as..."         , self.on_save_as     , pos = 5, accel = "C-S-S")
    self.save_as_songbook_menu = self.add_to_menu(file_menu, 0, "Save songbook as...", self.on_save_as_songbook, pos = 6)
    self.add_separator_to_menu(file_menu, pos = 7)
    
    self.import_menu          = self.add_to_menu(file_menu, 1, "Import", pos = 8)
    self.export_menu          = self.add_to_menu(file_menu, 1, "Export", pos = 9)
    self.export_songbook_menu = self.add_to_menu(file_menu, 1, "Export songbook", pos = 10)
    
    self.add_to_menu(file_menu, 0, "preview_print", self.on_preview_print, pos = 11)
    self.preview_print_songbook_menu = self.add_to_menu(file_menu, 0, "preview_print_songbook", self.on_preview_print_songbook, pos = 12)
    self.add_separator_to_menu(file_menu, pos = 13)
    #self.add_to_menu(file_menu, 0, "Close", self.on_close)
    
    self.add_separator_to_menu(file_menu)
    for file in globdef.config.PREVIOUS_FILES:
      self.add_to_menu(file_menu, 0, file, lambda obj, file = file: self.open_filename(file))
      
    #edit_menu = self.add_to_menu(self.menubar, 1, "Edit", self.on_edit_menu)
    edit_menu = self.edit_menu
    #self.undo_menu = self.add_to_menu(edit_menu, 0, "Undo", self.on_undo, accel = "C-Z")
    #self.redo_menu = self.add_to_menu(edit_menu, 0, "Redo", self.on_redo, accel = "C-Y")
    self.add_separator_to_menu(edit_menu)
    self.add_to_menu(edit_menu, 0, "Select all"              , self.on_select_all, accel = "C-A", accel_enabled = 0) # Disabled these accels,
    self.add_to_menu(edit_menu, 0, "Cut"                     , self.on_note_cut  , accel = "C-X", accel_enabled = 0) # because they block C-A, C-V,... in lyrics editor
    self.add_to_menu(edit_menu, 0, "Copy"                    , self.on_note_copy , accel = "C-C", accel_enabled = 0)
    self.add_to_menu(edit_menu, 0, "Paste"                   , self.on_note_paste, accel = "C-V", accel_enabled = 0)
    self.add_to_menu(edit_menu, 0, "Insert/remove beats..."  , self.on_insert_times)
    self.add_to_menu(edit_menu, 0, "Insert/remove bars..."   , self.on_insert_bars)
    self.add_separator_to_menu(edit_menu)
    self.add_to_menu(edit_menu, 0, "Rhythm and bars..."     , self.on_bars_prop)
    self.add_to_menu(edit_menu, 0, "Repeats and playlist...", self.on_playlist_prop)
    self.add_separator_to_menu(edit_menu)
    self.add_to_menu(edit_menu, 0, "Preferences...", self.on_preferences)
    
    play_menu = self.add_to_menu(self.menubar, 1, "Play")
    self.add_to_menu(play_menu, 0, "Play from the beginning"        , self.on_play)
    self.add_to_menu(play_menu, 0, "Play from here"                 , self.on_play_from_here, accel = " ")
    self.add_to_menu(play_menu, 0, "Play from the beginning in loop", self.on_play_in_loop)
    self.add_to_menu(play_menu, 0, "Play from here in loop"         , self.on_play_from_here_in_loop)
    self.add_to_menu(play_menu, 0, "Stop playing"                   , self.on_stop_playing)
    
    self.book_menu = self.add_to_menu(self.menubar, 1, "Songbook")
    self.add_to_menu(self.book_menu, 0, "New songbook"                      , self.on_new_songbook)
    self.add_separator_to_menu(self.book_menu)
    self.add_to_menu(self.book_menu, 0, "Songbook properties..."            , self.on_songbook_prop)
    self.add_separator_to_menu(self.book_menu)
    #self.no_songbook_menu = self.add_to_menu(book_menu, 0, "(no songbook opened)")
    #self.set_menu_enable(self.no_songbook_menu, False)
    self.songbook_songs_menu = []
    
    song_menu = self.add_to_menu(self.menubar, 1, "Song")
    self.add_to_menu(song_menu, 0, "New song"                          , self.on_new)
    self.add_separator_to_menu(song_menu)
    self.add_to_menu(song_menu, 0, "Song and instruments properties...", self.on_song_prop)
    
    instru_menu = self.add_to_menu(self.menubar, 1, "Instrument")
    self.add_to_menu(instru_menu, 0, "Add..."       , self.on_add_instrument)
    self.add_to_menu(instru_menu, 0, "Duplicate"    , self.on_dupplicate_instrument)
    self.add_to_menu(instru_menu, 0, "Remove"       , self.on_remove_instrument)
    self.add_to_menu(instru_menu, 0, "Move up"      , self.on_move_instrument_up)
    self.add_to_menu(instru_menu, 0, "Move down"    , self.on_move_instrument_down)
    self.add_separator_to_menu(instru_menu)
    self.add_to_menu(instru_menu, 0, "Instrument properties..."      , self.on_instrument_prop)
    
    note_menu = self.add_to_menu(self.menubar, 1, "Note")
    self.add_to_menu(note_menu, 0, "One octavo above", lambda *args: self.canvas and self.canvas.shift_selections_values( 12))
    self.add_to_menu(note_menu, 0, "One pitch above" , lambda *args: self.canvas and self.canvas.shift_selections_values( 1), accel = "+", accel_enabled = 0)
    self.add_to_menu(note_menu, 0, "One pitch below" , lambda *args: self.canvas and self.canvas.shift_selections_values(-1), accel = "-", accel_enabled = 0)
    self.add_to_menu(note_menu, 0, "One octavo below", lambda *args: self.canvas and self.canvas.shift_selections_values(-12))
    self.add_separator_to_menu(note_menu)
    self.add_to_menu(note_menu, 0, "Delete"                  , lambda *args: self.canvas and self.canvas.delete_notes(self.canvas.selections), accel = "Del", accel_enabled = 0)
    self.add_to_menu(note_menu, 0, "Arrange notes at fret...", self.on_note_arrange)
    self.add_to_menu(note_menu, 0, "Note properties..."                 , self.on_note_prop)
    
    for duration in [384, 192, 96, 48, 24, 12]:
      def f(event, duration = duration):
        if self.updating_menu or not self.canvas: return
        if self.canvas.selections:
          self.canvas.current_duration = duration
          notes = list(self.canvas.selections)
          old_durations = [note.duration for note in notes]
          
          def do_it(notes = notes):
            for note in notes:
              dotted  = note.is_dotted()
              triplet = note.is_triplet()
              note.duration = duration
              if dotted : note.duration *= 1.5
              if triplet: note.duration  = (note.duration * 2) // 3
              
          def undo_it(notes = notes):
            for i in range(len(notes)): notes[i] = old_durations[i]
            
          editobj3.undoredo.UndoableOperation(do_it, undo_it, _("change of %s") % _("duration"), self.undo_stack)
          
      setattr(self, "set_duration_%s" % duration, f)
    
    def toggle_dotted(*args):
      if self.canvas and not self.updating_menu: self.canvas.toggle_dotted()
    def toggle_triplet(*args):
      if self.canvas and not self.updating_menu: self.canvas.toggle_triplet()
    def toggle_accent(*args):
      if self.canvas and not self.updating_menu: self.canvas.toggle_accent()
    self.toggle_dotted  = toggle_dotted
    self.toggle_triplet = toggle_triplet
    self.toggle_accent  = toggle_accent 
    
    def on_duration_menu_cliked(*args):
      self.updating_menu = True
      if self.canvas and self.canvas.selections:
        note = tuple(self.canvas.selections)[0]
        self.set_menu_checked(dotted , note.is_dotted ())
        self.set_menu_checked(triplet, note.is_triplet())
      self.updating_menu = False
        
    duration_menu = self.add_to_menu(self.menubar, 1, "Duration", on_duration_menu_cliked)
    self.add_to_menu(duration_menu, 0, "Whole"        , self.set_duration_384, image = self.menu_image("note_384.png"))
    self.add_to_menu(duration_menu, 0, "Half"         , self.set_duration_192, image = self.menu_image("note_192.png"))
    self.add_to_menu(duration_menu, 0, "Quarter"      , self.set_duration_96 , image = self.menu_image("note_96.png"))
    self.add_to_menu(duration_menu, 0, "Eighth"       , self.set_duration_48 , image = self.menu_image("note_48.png"))
    self.add_to_menu(duration_menu, 0, "Sixteenth"    , self.set_duration_24 , image = self.menu_image("note_24.png"))
    self.add_to_menu(duration_menu, 0, "Thirty-second", self.set_duration_12 , image = self.menu_image("note_12.png"))
    self.add_separator_to_menu(duration_menu)
    dotted  = self.add_to_menu(duration_menu, 0, "Dotted"       , toggle_dotted , type = "check", accel = ".", accel_enabled = 0)
    triplet = self.add_to_menu(duration_menu, 0, "Triplet"      , toggle_triplet, type = "check")
    self.add_separator_to_menu(duration_menu)
    self.add_to_menu(duration_menu, 0, "Longer"       , self.set_duration_longer , accel = "*")
    self.add_to_menu(duration_menu, 0, "Shorter"      , self.set_duration_shorter, accel = "/")
    
    
    def on_volume_fx_menu_cliked(*args):
      self.updating_menu = True
      if self.canvas and self.canvas.selections:
        note = tuple(self.canvas.selections)[0]
        if   note.volume == 255:       self.set_menu_checked(volume_menus[0], True)
        else:                          self.set_menu_checked(volume_menus[0], False)
        if   190 < note.volume  < 255: self.set_menu_checked(volume_menus[1], True)
        else:                          self.set_menu_checked(volume_menus[1], False)
        if   100 < note.volume <= 190: self.set_menu_checked(volume_menus[2], True)
        else:                          self.set_menu_checked(volume_menus[2], False)
        if   note.volume <= 100:       self.set_menu_checked(volume_menus[3], True)
        else:                          self.set_menu_checked(volume_menus[3], False)
        
        fxs = set([note.link_fx, note.duration_fx, note.strum_dir_fx])
        if note.fx == "bend": fxs.add("bend%s" % note.bend_pitch)
        else:                  fxs.add(note.fx)
        for key, menu in fx_menus.items():
          self.set_menu_checked(menu, key in fxs)
      self.updating_menu = False

    for volume in [255, 204, 120, 50]:
      def f(event, volume = volume):
        if self.canvas and not self.updating_menu: self.canvas.set_selections_volume(volume)
      setattr(self, "set_volume_%s" %  volume, f)
      
    for type, fxs in model.ALL_FXS:
      if type: type += "_"
      for fx in fxs:
        if not fx: continue
        for arg in model.FXS_ARGS.get(fx, [None]):
          def f(event, type = type, fx = fx, arg = arg):
            if self.canvas and not self.updating_menu: getattr(self.canvas, "set_selections_%sfx" % type)(fx, arg)
          if arg: setattr(self, "set_fx_%s%s" % (fx, str(arg).replace(".", "")), f)
          else:   setattr(self, "set_fx_%s"   %  fx, f)
          
    effect_menu = self.add_to_menu(self.menubar, 1, "Effect", on_volume_fx_menu_cliked)
    volume_menus = [
      self.add_to_menu(effect_menu, 0, "Accentuated"            , self.set_volume_255, type = "radio", accel = "Enter", accel_enabled = 0),
      self.add_to_menu(effect_menu, 0, "Normal"                 , self.set_volume_204, type = "radio"),
      self.add_to_menu(effect_menu, 0, "Low"                    , self.set_volume_120, type = "radio"),
      self.add_to_menu(effect_menu, 0, "Very low"               , self.set_volume_50 , type = "radio"),
      ]
    self.add_separator_to_menu(effect_menu)
    self.add_to_menu(effect_menu, 0, "Remove all effects"       , self.set_fx_none   , accel = "n", accel_enabled = 0)
    
    
    self.add_separator_to_menu(effect_menu)
    fx_menus = self.fx_menus = {
      "bend0.5" : self.add_to_menu(effect_menu, 0, "bend 0.5" , self.set_fx_bend05  , type = "check", accel = "b", accel_enabled = 0),
      "bend1.0" : self.add_to_menu(effect_menu, 0, "bend 1"   , self.set_fx_bend10  , type = "check"),
      "bend1.5" : self.add_to_menu(effect_menu, 0, "bend 1.5" , self.set_fx_bend15  , type = "check"),
      "tremolo" : self.add_to_menu(effect_menu, 0, "tremolo"  , self.set_fx_tremolo , type = "check", accel = "t", accel_enabled = 0),
      "dead"    : self.add_to_menu(effect_menu, 0, "dead note", self.set_fx_dead    , type = "check", accel = "d", accel_enabled = 0),
      "roll"    : self.add_to_menu(effect_menu, 0, "roll"     , self.set_fx_roll    , type = "check", accel = "r", accel_enabled = 0),
      "harmonic": self.add_to_menu(effect_menu, 0, "harmonic" , self.set_fx_harmonic, type = "check", accel = "h", accel_enabled = 0),
      }
    self.add_separator_to_menu(effect_menu)
    fx_menus.update({
      "appoggiatura": self.add_to_menu(effect_menu, 0, "appoggiatura", self.set_fx_appoggiatura, type = "check", accel = "a", accel_enabled = 0),
      "fermata"     : self.add_to_menu(effect_menu, 0, "fermata"     , self.set_fx_fermata     , type = "check", accel = "p", accel_enabled = 0),
      "breath"      : self.add_to_menu(effect_menu, 0, "breath"      , self.set_fx_breath      , type = "check"),
      })
    self.add_separator_to_menu(effect_menu)
    fx_menus.update({
      "link"  : self.add_to_menu(effect_menu, 0, "link" , self.set_fx_link   , type = "check", accel = "l", accel_enabled = 0),
      "slide" : self.add_to_menu(effect_menu, 0, "slide", self.set_fx_slide  , type = "check", accel = "s", accel_enabled = 0),
    })
    self.add_separator_to_menu(effect_menu)
    fx_menus.update({
      "up"         : self.add_to_menu(effect_menu, 0, "up"        , self.set_fx_up        , type = "check", accel = "^", accel_enabled = 0),
      "down"       : self.add_to_menu(effect_menu, 0, "down"      , self.set_fx_down      , type = "check", accel = "_", accel_enabled = 0),
      "down_thumb" : self.add_to_menu(effect_menu, 0, "down_thumb", self.set_fx_down_thumb, type = "check", accel = ",", accel_enabled = 0),
    })
    
    help_menu = self.add_to_menu(self.menubar, 1, "Help")
    self.add_to_menu(help_menu, 0, "About..."         , self.on_about)
    self.add_to_menu(help_menu, 0, "Manual..."        , self.on_manual, accel = "F1")
    self.add_separator_to_menu(help_menu)
    self.add_to_menu(help_menu, 0, "Dump"             , self.on_dump)
    self.add_to_menu(help_menu, 0, "GC"               , self.on_gc)
    
    
    songwrite3.plugins.create_plugins_ui(self)
    
    self.song     = None
    self.songbook = None
    self.instrument_chooser_pane = None
    self.paned    = None
    self.scrolled = None
    self.canvas   = None
    self.selected_partition = None
    self.selected_mesure    = None
    self.selected_note      = None
    
    if   isinstance(song, model.Song)    : self.set_songbook(None); self.set_song(song)
    elif isinstance(song, model.Songbook): self.set_songbook(song)
    else:                                  self.set_song    (None)
    
    self.window.closeEvent = self.closeEvent
    self.destroy           = self.window.destroy
    
    toolbar = self.toolbar = qtwidgets.QToolBar()
    self.add_to_menu(toolbar, 0, "Save",           self.on_save)
    self.add_to_menu(toolbar, 0, "preview_print",  self.on_preview_print)
    self.add_to_menu(toolbar, 0, "Play from here", self.on_play_from_here)
    self.add_to_menu(toolbar, 0, "Stop playing",   self.on_stop_playing)
    
    self.toolbar.addSeparator()
    
    radio_group = qtwidgets.QActionGroup(self.window)
    for duration in [384, 192, 96, 48, 24, 12]:
      action = self.add_to_menu(toolbar, 0, model.DURATIONS[duration], getattr(self, "set_duration_%s" % duration), image = self.menu_image("note_%s.png" % duration))
      action.setCheckable(True)
      radio_group.addAction(action)
      setattr(self, "toolbar_duration_%s" % duration, action)
    self.toolbar_dotted  = self.add_to_menu(toolbar, 0, "Dotted" , self.toggle_dotted , image = self.menu_image("note_144.png"))
    self.toolbar_triplet = self.add_to_menu(toolbar, 0, "Triplet", self.toggle_triplet, image = self.menu_image("note_64.png" ))
    self.toolbar_dotted .setCheckable(True)
    self.toolbar_triplet.setCheckable(True)
    
    self.toolbar.addSeparator()
    
    self.toolbar_accent = self.add_to_menu(toolbar, 0, "Accentuated", self.toggle_accent, image = self.menu_image("effet_accent.png" ))
    self.toolbar_accent.setCheckable(True)
    
    self.toolbar_buttons = {}
    for type, fxs in model.ALL_FXS:
      for fx in fxs:
        if not fx: continue
        if fx == "bend": func = self.set_fx_bend05
        else:            func = getattr(self, "set_fx_%s" % fx)
        self.toolbar_buttons[fx] = action = self.add_to_menu(toolbar, 0, fx, func, image = self.menu_image("effet_%s.png" % fx))
        action.setCheckable(True)
        
    zoom_levels = [0.25, 0.35, 0.5, 0.75, 1.0, 1.5, 2.0, 3.0, 4.0]
    
    def zoom_in (menu_item):
      if self.canvas: self.canvas.change_zoom( 1)
    def zoom_out(menu_item):
      if self.canvas: self.canvas.change_zoom(-1)
    def set_zoom(zoom):
      if self.canvas:
        zoom = int(zoom.replace("%", "")) / 100.0
        self.canvas.set_zoom(zoom)
    self.on_set_zoom = set_zoom
    
    zoom_toolbar = qtwidgets.QToolBar()
    zoom_out_menu = self.add_to_menu(zoom_toolbar, 0, " ‒ ", zoom_out)
    
    self.zoom_menu  = qtwidgets.QComboBox()
    for zoom in zoom_levels:
      self.zoom_menu.addItem("%s%%" % int(zoom * 100.0), zoom)
    self.zoom_menu.setFrame(False)
    self.zoom_menu.setCurrentText("100%")
    self.zoom_menu.currentTextChanged.connect(set_zoom)
    zoom_toolbar.addWidget(self.zoom_menu)
    zoom_in_menu  = self.add_to_menu(zoom_toolbar, 0, " + ", zoom_in)
    
    self.toolbar_layout = qtwidgets.QHBoxLayout()
    self.toolbar_layout.addWidget(toolbar, 1)
    self.toolbar_layout.addWidget(zoom_toolbar)
    self.layout.insertLayout(1, self.toolbar_layout)
    
    if globdef.config.WINDOW_WIDTH == 0:
      self.window.showMaximized()
    else:
      self.window.resize(globdef.config.WINDOW_WIDTH, globdef.config.WINDOW_HEIGHT)
      self.window.showNormal()
      
    if globdef.config.WINDOW_X != -1:
      self.window.move(globdef.config.WINDOW_X, globdef.config.WINDOW_Y)
      
  def create_content(self):
    self.window.setWindowTitle("Songwrite3")
    self.instrument_chooser_pane = InstrumentChooserPane(self)
    self.window.layout().addWidget(self.instrument_chooser_pane)
    
  def menu_image(self, filename):
    icon = qtgui.QIcon(os.path.join(globdef.DATADIR, filename))
    return icon
  
  def set_fx_none(self, *args):    self.canvas.remove_selections_fx()
  
  def set_selected_partition(self, partition):
    self.selected_partition = partition
    if self.partition_editor and partition:
      if self.partition_editor.window.isVisible(): self.partition_editor.edit(partition)
      else:                                        self.partition_editor = None
    
  def set_selected_mesure(self, mesure):
    self.selected_mesure = mesure
    if self.mesure_editor and mesure:
      if self.mesure_editor.window.isVisible(): self.mesure_editor.edit(mesure)
      else:                                     self.mesure_editor = None
      
  def set_selected_note(self, note):
    self.updating_menu = True
    self.selected_note = note
    if self.note_editor and note:
      if self.note_editor.window.isVisible(): self.note_editor.edit(note)
      else:                                   self.note_editor = None
    if isinstance(note, introsp.ObjectPack):
      mesure_ids = set([self.song.mesure_at(n).get_number() + 1 for n in note.objects])
      if len(mesure_ids) == 1:
        mesure_label = self.song.mesure_at(note.objects[0])
      else:
        mesure_label = _("Bars #%s to #%s") % (min(mesure_ids), max(mesure_ids))
      if len(note.objects) == 2:
        delta = abs(note.objects[0].value - note.objects[1].value)
        note_label = _("__interval__%s" % delta)
        if note_label.startswith("_"): note_label = "2 %s" % _("notes")
      else:
        note_label = "%s %s" % (len(note.objects), _("notes"))
      note = note.objects[-1]
    else:
      mesure_label = self.song.mesure_at(note)
      note_label = str(note)
      
    current_duration = note.base_duration()
    self.set_menu_checked(getattr(self, "toolbar_duration_%s" % current_duration), True)
    self.set_menu_checked(self.toolbar_dotted , note.is_dotted())
    self.set_menu_checked(self.toolbar_triplet, note.is_triplet())
    self.set_menu_checked(self.toolbar_accent , note.volume == 255)
    
    fxs = { note.fx, note.link_fx, note.duration_fx, note.strum_dir_fx }
    for key, button in self.toolbar_buttons.items(): self.set_menu_checked(button, key in fxs)
    
    self.window.setWindowTitle("Songwrite3 -- %s -- %s -- %s" % (self.song.title, mesure_label, note_label))
    self.updating_menu = False
      
  def get_selected_notes(self):
    if isinstance(self.selected_note, introsp.ObjectPack): return self.selected_note.objects
    return [self.selected_note]
  
  def partition_categories(self):
    next_strophe_number = len([lyrics for lyrics in self.song.partitions if isinstance(lyrics, model.Lyrics) and _("Strophe #%s").replace("%s", "") in lyrics.header]) + 1
    
    return PartitionCategory(_("Available score views"), _("Select the score and score view you want to add."),
        [model.Partition(None, view_type.default_instrument, view_type) for view_type in [model.GuitarView, model.BassView]] +
        [PartitionCategory(_("More tablatures..."), _("Tablature (for guitar-like instruments)"),
          [model.Partition(None, view_type.default_instrument, view_type) for view_type in model.VIEWS["tab"] if not view_type is model.GuitarView and not view_type is model.BassView])
        ] +
        [model.Partition(None, view_type.default_instrument, view_type) for view_type in [model.PianoView, model.GuitarStaffView, model.VocalsView]] +
        [PartitionCategory(_("More staffs..."), _("Staff (for any instrument)"),
          [model.Partition(None, view_type.default_instrument, view_type) for view_type in model.VIEWS["staff"] if not view_type is model.PianoView and not view_type is model.VocalsView and not view_type is model.GuitarStaffView])
        ] +
        [model.Partition(None, model.TomView.default_instrument, model.TomView)] +
        [PartitionCategory(_("More drums..."), _("Drum tablatures (for drums)"),
          [model.Partition(None, view_type.default_instrument, view_type) for view_type in model.VIEWS["drums"] if not view_type is model.TomView])
        ] +
        [
        model.Lyrics(None, header = _("Chorus")),
        model.Lyrics(None, header = _("Strophe #%s") % next_strophe_number),
        ])

  def edit(self, o, *args, **kargs):
    return editobj3.edit(o, undo_stack = self.undo_stack, master = self, menubar = False, *args, **kargs)
  
  def on_new_app(self, *args): self.__class__()
  
  def close_dialog(self, *args): self.closeEvent(None)
    
  def closeEvent(self, e):
    if self.check_save():
      if e: e.ignore()
      return
    if e: e.accept()
    
    if self.song: del APPS[self.song]
    if self.songbook: del APPS[self.songbook]
    self.destroy()
    
    if len(APPS) == 0:
      globdef.config.WINDOW_X = self.window.x()
      globdef.config.WINDOW_Y = self.window.y()
      if self.window.isMaximized():
        globdef.config.WINDOW_WIDTH  = globdef.config.WINDOW_HEIGHT = 0
      else:
        globdef.config.WINDOW_WIDTH  = self.window.width()
        globdef.config.WINDOW_HEIGHT = self.window.height()
      globdef.config.save()
      sys.exit()

      
  def prompt_open_filename(self, exts = [".sw.xml"], typename = "", title = None):
    ext_label = self.get_ext_label(exts, typename)
    if self.filename: dirname = self.filename
    else:             dirname = ""
    filename = qtwidgets.QFileDialog.getOpenFileName(self.window, title or _("Open..."), dirname, ext_label)[0]
    return filename
  
  def prompt_save_filename(self, exts = [".sw.xml"], typename = "", title = None):
    ext_label = self.get_ext_label(exts, typename)
    if self.filename:
      filename  = self.filename.split(".")[0] + exts[0]
    else:
      filename  = remove_accents(self.song.title.lower().replace(" ", "_").replace("'", "").replace("/", "_")) + exts[0]
      
    filename = qtwidgets.QFileDialog.getSaveFileName(self.window, title or _("Save as..."), filename, ext_label)[0]
    if filename:
      for ext in exts:
        if filename.endswith(ext): break
      else:
        filename += exts[0]
    return filename
  
  def get_ext_label(self, exts, typename):
    if exts == [".sw.xml"]:
      ext_label = _("__songwrite_file_format__") + " (*.sw.xml)"
    else:
      exts_joined = " ".join("*%s" % ext for ext in exts)
      typename    = typename or exts_joined
      ext_label   = "%s %s (%s)" % (_("Files"), typename, exts_joined)
    return ext_label + ";;%s (*.*)" % _("__all_files__")

  
  def on_open(self, event = None):
    if self.check_save(): return
    
    filename = self.prompt_open_filename()
    if filename:
      self.open_filename(filename, do_check_save = 0) # check_save has already been done before asking the filename
      
  def check_save(self):
    if not self.song: return False
    
    if self.undo_stack.undoables == self.last_undoables: return 0 # The undoable actions have not changed => nothing to save.
    msg = qtwidgets.QMessageBox()
    msg.setText(_("Save modifications before closing?"))
    #msg.setStandardButtons(qtwidgets.QMessageBox.Save | qtwidgets.QMessageBox.Discard | qtwidgets.QMessageBox.Cancel)
    #msg.setDefaultButton(qtwidgets.QMessageBox.Cancel)
    save_button    = msg.addButton(_("Save"), 0)
    discard_button = msg.addButton(_("Close without saving"), 0)
    cancel_button  = msg.addButton(_("Cancel"), 0)
    msg.setDefaultButton(cancel_button)
    msg.setEscapeButton(cancel_button)
    ret = msg.exec()
    if   ret == 1: return False
    elif ret == 2: return True
    elif ret == 0:
      self.on_save()
      return self.check_save() # The user may have canceled the "save as" dialog box !
    
  def set_song(self, song):
    if self.song:
      del APPS[self.song]
      if self.filename: model.unload_song(self.filename)
      
    self.undo_stack.clear()
    self.last_undoables = []
    self.song = song
    
    if self.instrument_chooser_pane:
      self.instrument_chooser_pane.hide()
      self.window.layout().addWidget(self.instrument_chooser_pane)
      self.instrument_chooser_pane.destroy()
      
    if self.canvas:
      zoom = self.canvas.zoom
      self.canvas.hide()
      self.window.layout().addWidget(self.canvas)
      self.canvas.destroy()
    else:
      zoom = 1.0
      
    if song:
      APPS[song] = self
      self.filename = song.filename
      self.window.setWindowTitle("Songwrite3 -- %s" % song.title)
      self.canvas = Canvas(self, self.song, zoom)
      self.window.layout().addWidget(self.canvas)
      
    else:
      self.window.setWindowTitle("Songwrite3")
      self.instrument_chooser_pane = InstrumentChooserPane(self)
      self.window.layout().addWidget(self.instrument_chooser_pane)
      
    
  def set_songbook(self, songbook):
    if self.songbook:
      del APPS[self.songbook]
      unobserve(self.songbook.song_refs, self.on_songbook_changed)
      
    self.songbook = songbook
    
    if self.songbook:
      APPS[self.songbook] = self
      if self.songbook.song_refs: self.set_song(self.songbook.song_refs[0].get_song())
      observe(self.songbook.song_refs, self.on_songbook_changed)
      
    self.rebuild_songbook_menu()
    
  def rebuild_songbook_menu(self):
    for menu in self.songbook_songs_menu: self.book_menu.removeAction(menu)
    
    if self.songbook:
      self.songbook_songs_menu = []
      for song_ref in self.songbook.song_refs:
        def open_song(event, song_ref = song_ref):
          self.set_song(song_ref.get_song())
        menu = self.add_to_menu(self.book_menu, 0, song_ref.title, open_song)
        self.songbook_songs_menu.append(menu)
        
    else:
      no_songbook_menu = self.add_to_menu(self.book_menu, 0, "(no songbook opened)")
      self.set_menu_enable(no_songbook_menu, False)
      self.songbook_songs_menu = [no_songbook_menu]
      
  def on_songbook_changed(self, obj, type, new, old):
    self.rebuild_songbook_menu()
    
    
  def on_file_menu(self, *args):
    if self.songbook:
      self.set_menu_enable(self.save_as_songbook_menu      , 1)
      self.set_menu_enable(self.preview_print_songbook_menu, 1)
      self.set_menu_enable(self.export_songbook_menu       , 1)
    else:
      self.set_menu_enable(self.save_as_songbook_menu      , 0)
      self.set_menu_enable(self.preview_print_songbook_menu, 0)
      self.set_menu_enable(self.export_songbook_menu       , 0)
    
  def on_new(self, *args):
    if self.check_save(): return
    
    self.filename = None
    self.set_songbook(None)
    self.set_song    (None)
    
  def on_new_songbook(self, *args):
    if self.check_save(): return

    songbook = model.Songbook()
    songbook.authors   = getpass.getuser().title()
    songbook.title     = _("%s's songbook") % songbook.authors
    
    self.set_songbook(songbook)
    self.on_songbook_prop()
    
  def on_open(self, event = None):
    if self.check_save(): return
    
    filename = self.prompt_open_filename()
    if not filename: return
    self.open_filename(filename, do_check_save = 0) # check_save has already been done before asking the filename
    
  def open_filename(self, filename, do_check_save = 1):
    globdef.config.LAST_DIR = os.path.dirname(filename)
    
    if do_check_save and self.check_save(): return
    
    if not os.path.exists(filename):
      editobj3.edit(_("File does not exist!"), on_validate = lambda o: None)
    else:
      self.filename = filename
      song = model.get_song(filename)
      if isinstance(song, model.Song):
        self.set_songbook(None)
        self.set_song    (song)
      else:
        self.set_songbook(song)
        self.on_songbook_prop()
      globdef.config.add_previous_file(filename)
      
  def on_save_as(self, event = None):
    filename = self.prompt_save_filename()
    if not filename: return
    self.filename = self.song.filename = filename
    self.on_save(save_song = 1, save_songbook = 0)
    
  def on_save_as_songbook(self, event = None):
    filename = self.prompt_save_filename()
    if not filename: return
    if not filename.endswith(".sw.xml"): filename += ".sw.xml"
    self.songbook.set_filename(filename)
    self.on_save(save_song = 0, save_songbook = 1)
    
  def on_save(self, event = None, save_song = 1, save_songbook = 1):
    if not self.song: return
    if save_song:
      ok_song = 0
      
      if not self.filename: self.on_save_as()
      if self.filename:
        #self.song.__xml__(codecs.lookup("utf8")[3](open(self.filename, "w")))
        s = StringIO()
        self.song.__xml__(s)
        s = s.getvalue()
        open(self.filename, "w").write(s)
        ok_song = 1
        
        if not self.songbook: globdef.config.add_previous_file(self.filename)
        
    else: ok_song = 1
    
    
    if save_songbook and self.songbook:
      ok_songbook = 0
      
      if not self.songbook.filename: self.on_save_as_songbook()
      if self.songbook.filename:
        #self.songbook.__xml__(codecs.lookup("utf8")[3](open(self.songbook.filename, "w")))
        s = StringIO()
        self.songbook.__xml__(s)
        s = s.getvalue()
        open(self.songbook.filename, "w").write(s)
        ok_songbook = 1
        
        globdef.config.add_previous_file(self.songbook.filename)
        
    else: ok_songbook = 1
    
    
    if ok_song and ok_songbook: self.last_undoables = self.undo_stack.undoables[:]
    
  def on_preview_print(self, event = None):
    if not self.song: return
    try:
      songwrite3.plugins.get_exporter("pdf").print_preview(self.song)
    except:
      self._treat_export_error()
      
  def on_preview_print_songbook(self, event = None):
    if not self.songbook: return
    try:
      songwrite3.plugins.get_exporter("pdf").print_preview(self.songbook)
    except:
      self._treat_export_error()
      
  def _treat_export_error(self):
    error_class, error, trace = sys.exc_info()
    sys.excepthook(error_class, error, trace)
    print(file = sys.stderr)
    
    if isinstance(error, model.SongwriteError):
      editobj3.edit(error)
      if isinstance(error, model.TimeError):
        self.canvas.deselect_all()
        if   error.note:
          self.canvas.partition_2_drawer[error.note.partition].select_note(error.note)
        elif error.partition and error.time:
          self.canvas.partition_2_drawer[error.partition].select_at(error.time, 0, 0)
        elif error.time:
          self.canvas.partition_2_drawer[self.song.partitions[0]].select_at(error.time, 0, 0)
    else:
      editobj3.edit(_(error.__class__.__name__) + "\n\n" + str(error))
      
      
  def on_about(self, *args):
    class About(object):
      def __init__(self):
        self.details       = _("__about__")
        self.icon_filename = os.path.join(globdef.DATADIR, "songwrite_about.png")
        self.url           = "http://www.lesfleursdunormal.fr/static/informatique/songwrite/index_en.html"
        self.licence       = "GPL"
        self.authors       = "Jean-Baptiste 'Jiba' Lamy <jibalamy@free.fr>"
        
      def __str__(self): return "Songwrite 3 version %s" % model.VERSION
      
    descr = introsp.description(About)
    descr.def_attr("details"      , field.HiddenField)
    descr.def_attr("icon_filename", field.HiddenField)
    self.edit(About())
    
  def on_manual(self, *args):
    DOCDIR = os.path.join(globdef.APPDIR, "doc")
    if not os.path.exists(DOCDIR):
      import glob
      DOCDIR = glob.glob(os.path.join(globdef.APPDIR, "..", "doc", "songwrite3*"))[0]
      
      if not os.path.exists(DOCDIR):
        DOCDIR = glob.glob("/usr/share/doc/songwrite3*")[0]
      
        if not os.path.exists(DOCDIR):
          DOCDIR = glob.glob("/usr/share/local/doc/songwrite3*")[0]
          
          if not os.path.exists(DOCDIR):
            DOCDIR = glob.glob("/usr/doc/songwrite3")[0]
            
    DOC = os.path.join(DOCDIR, locale.getdefaultlocale()[0][:2], "doc.pdf")
    if not os.path.exists(DOC):
      DOC = os.path.join(DOCDIR, "en", "doc.pdf")
      
    DOC = os.path.abspath(DOC)
    pdf_command = globdef.config.get_preview_command_pdf()
    if "%s" in pdf_command:
      #os.system(pdf_command % DOC)
      p = subprocess.Popen(pdf_command % DOC, shell = True, close_fds = True)
    else:
      if pdf_command.endswith("-"): pdf_command = pdf_command[-1]
      #os.system(pdf_command + " " + DOC)
      p = subprocess.Popen(pdf_command + " " + DOC, shell = True, close_fds = True)
      
  def on_add_instrument(self, event = None):
    if self.selected_partition:
      if isinstance(self.selected_partition, introsp.ObjectPack): selected_partition = self.selected_partition.objects[-1]
      else:                                                       selected_partition = self.selected_partition
    else:                                                         selected_partition = None
    def callback(*args): pass
    
    song = self.song or new_song()
    descr = introsp.description(model.Song)
    if selected_partition:
      action = introsp.InsertAction(descr.attributes["partitions"])
      new_child = descr.do_action(action, self.undo_stack, None, song, descr.attributes["partitions"], selected_partition, callback)
    else:
      action = introsp.AppendAction(descr.attributes["partitions"])
      new_child = descr.do_action(action, self.undo_stack, None, song, callback)
    if not self.song is song: self.set_song(song)
    
  def on_dupplicate_instrument(self, event = None):
    if not self.song: return
    if isinstance(self.selected_partition, introsp.ObjectPack):
      partitions = self.selected_partition.objects
    else: partitions = [self.selected_partition]
    
    insert_at = max([self.song.partitions.index(partition) for partition in partitions])
    
    xml = StringIO()
    xml.write("<song>\n")
    context = model._XMLContext()
    for partition in partitions: partition.__xml__(xml, context)
    xml.write("</song>")
    
    import songwrite3.stemml as stemml
    xml.seek(0)
    song = stemml.parse(xml)
    
    def do_it(song = song):
      for partition in song.partitions:
        self.song.insert_partition(insert_at, partition)
        partition.song = self.song
    def undo_it(song = song):
      for partition in song.partitions:
        self.song.remove_partition(partition)
    undoredo.UndoableOperation(do_it, undo_it, _("Duplicate instrument"), self.undo_stack)
    
  def on_remove_instrument(self, event = None):
    if self.song and self.selected_partition:
      descr = introsp.description(self.song.__class__)
      descr.do_action(introsp._REMOVE_ACTION, self.undo_stack, None, self.song, descr.attributes["partitions"], self.selected_partition)
      self.set_selected_partition(None)
      
  def on_move_instrument_up(self, event = None):
    if self.song and self.selected_partition:
      descr = introsp.description(self.song.__class__)
      descr.do_action(introsp._MOVE_UP_ACTION, self.undo_stack, None, self.song, descr.attributes["partitions"], self.selected_partition)
      
  def on_move_instrument_down(self, event = None):
    if self.song and self.selected_partition:
      descr = introsp.description(self.song.__class__)
      descr.do_action(introsp._MOVE_DOWN_ACTION, self.undo_stack, None, self.song, descr.attributes["partitions"], self.selected_partition)
    
  def on_song_prop(self, event = None):
    if self.song: self.edit(self.song)
    
  def on_instrument_prop(self, event = None):
    if self.song and self.selected_partition:
      self.partition_editor = self.edit(self.selected_partition)
      self.partition_editor.window.resize(self.partition_editor.window.width(), app.desktop().screenGeometry(self.window).height() * 0.8)
     
  def on_note_prop(self, event = None):
    if self.selected_note: self.note_editor = self.edit(self.selected_note)
    
  def on_songbook_prop(self, event = None):
    if not self.songbook:
      self.on_new_songbook()
      return
    
    def on_edit_child(song_ref):
      if isinstance(song_ref, model.SongRef):
        self.set_song(song_ref.get_song())
        
    self.edit(self.songbook, on_edit_child = on_edit_child)
    
  def on_bars_prop(self, event = None):
    mesures = self.canvas.get_selected_mesures()
    if mesures:
      if len(mesures) == 1: mesures = mesures[0]
      else:                 mesures = introsp.ObjectPack(mesures)
      self.mesure_editor = self.edit(mesures)
      
  def on_playlist_prop(self, event = None):
    if self.song:
      editor = self.edit(self.song.playlist)
      editor.window.resize(650, 400)
    
  def on_preferences(self, *args):
    globdef.config.edit()
    
  def on_play(self, *args):
    if self.canvas:
      player.play(songwrite3.midi.song_2_midi(self.song, trackable = 1), 0, self.canvas.play_tracker)
  def on_play_in_loop(self, *args):
    if self.canvas:
      player.play(songwrite3.midi.song_2_midi(self.song, trackable = 1), 1, self.canvas.play_tracker)
  def on_play_from_here(self, *args):
    if self.canvas:
      if player.is_playing(): player.stop()
      else:                   player.play(songwrite3.midi.song_2_midi(self.song, self.canvas.get_selected_time(), trackable = 1), 0, self.canvas.play_tracker)
  def on_play_from_here_in_loop(self, *args):
    if self.canvas:
      player.play(songwrite3.midi.song_2_midi(self.song, self.canvas.get_selected_time(), trackable = 1), 1, self.canvas.play_tracker)
  def on_stop_playing(self, *args):
    player.stop()
  
  def on_space(self, *args): # space : play from the current position
    if player.is_playing(): player.stop()
    else: self.on_play_from_here()
    
          
  def on_select_all(self, *args):
    if self.canvas: self.canvas.select_all()
    
  def on_note_cut(self, *args):
    if self.canvas: self.canvas.on_cut()
    
  def on_note_copy(self, *args):
    if self.canvas: self.canvas.on_copy()
    
  def on_note_paste(self, *args):
    if self.canvas: self.canvas.on_paste()
    
  def on_insert_times(self, *args):
    if not self.canvas: return
    class InsertTimesOptions(object):
      def __init__(self):
        self.nb_time       = 1
        self.details       = _("""Use negative values for deleting beats.""")
        self.icon_filename = os.path.join(globdef.DATADIR, "song.png")
      def __str__(self): return _("Insert/remove beats...")
    
    o = InsertTimesOptions()
    
    def on_validate(o):
      if (not o) or (not o.nb_time): return
      at = self.canvas.get_selected_time()
      delta = o.nb_time * 96
      removed = []
      
      def shift(delta):
        old_removed = removed[:]
        del removed[:]

        for partition in self.song.partitions:
          if isinstance(partition, model.Partition):
            i = 0
            while i < len(partition.notes):
              note = partition.notes[i]
              if note.time >= at:
                if (delta < 0) and note.time < at - delta:
                  partition.remove(partition.notes[i])
                  removed.append((partition, note))
                  continue
                else: note.time += delta
              i += 1
              
        for partition, note in old_removed: partition.add_note(note)
        
        self.song.rythm_changed()
        
        self.canvas.render_all()
        self.canvas.update_time()
        
      def do_it  (): shift( delta)
      def undo_it(): shift(-delta)

      if o.nb_time > 0: name = _("Insert %s beat(s)") %  o.nb_time
      else:             name = _("Delete %s beat(s)") % -o.nb_time
      undoredo.UndoableOperation(do_it, undo_it, name, self.undo_stack)
      
    editobj3.edit(o, on_validate = on_validate, menubar = False)
    
  def on_insert_bars(self, *args):
    if not self.canvas: return
    class InsertBarsOptions(object):
      def __init__(self):
        self.nb_bar        = 1
        self.details       = _("""Use negative values for deleting bars.""")
        self.icon_filename = os.path.join(globdef.DATADIR, "song.png")
      def __str__(self): return _("Insert/remove bars...")
    
    o = InsertBarsOptions()

    def on_validate(o):
      if (not o) or (not o.nb_bar): return
    
      mesures = self.canvas.get_selected_mesures()
      if mesures: mesure = mesures[0]
      else:       mesure = self.song.mesures[0]
      
      mesure_pos = self.song.mesures.index(mesure)
      at = mesure.time
      removed = []
      playlist_items_values = []
      
      def shift(nb_bar):
        # Add / remove mesures
        time = self.song.mesures[mesure_pos].time
        if nb_bar > 0:
          for i in range(mesure_pos, mesure_pos + nb_bar):
            self.song.mesures.insert(i, model.Mesure(self.song, time, mesure.tempo, mesure.rythm1, mesure.rythm2, mesure.syncope))
            time += mesure.duration
        else:
          del self.song.mesures[mesure_pos : mesure_pos - nb_bar]
          i = mesure_pos
          
        # Shift playlist
        if playlist_items_values:
          for item, from_mesure, to_mesure in playlist_items_values:
            item.from_mesure = from_mesure
            item.to_mesure   = to_mesure
          del playlist_items_values[:]
        else:
          for item in self.song.playlist.playlist_items:
            playlist_items_values.append((item, item.from_mesure, item.to_mesure))
            if   item.from_mesure >= mesure_pos:
              if item.from_mesure <  mesure_pos - nb_bar: item.from_mesure  = mesure_pos
              else:                                       item.from_mesure += nb_bar
            if   item.to_mesure   >= mesure_pos - 1:
              if item.to_mesure   <  mesure_pos-1-nb_bar: item.to_mesure   = mesure_pos - 1
              else:                                       item.to_mesure  += nb_bar
              
        # Shift notes
        old_removed = removed[:]
        del removed[:]

        delta = nb_bar * mesure.duration
        for partition in self.song.partitions:
          if isinstance(partition, model.Partition):
            i = 0
            while i < len(partition.notes):
              note = partition.notes[i]
              if note.time >= at:
                if (delta < 0) and note.time < at - delta:
                  partition.remove(partition.notes[i])
                  removed.append((partition, note))
                  continue
                else: note.time += delta
              i += 1
              
        for partition, note in old_removed: partition.add_note(note)

        self.song.rythm_changed() # AFTER moving the notes !!!
        
        self.canvas.render_all()
        self.canvas.update_time()
        
      def do_it  (): shift( o.nb_bar)
      def undo_it(): shift(-o.nb_bar)
      
      if o.nb_bar > 0: name = _("Insert %s bar(s)") %  o.nb_bar
      else:            name = _("Delete %s bar(s)") % -o.nb_bar
      undoredo.UndoableOperation(do_it, undo_it, name, self.undo_stack)
      
    editobj3.edit(o, on_validate = on_validate, menubar = False)
    
  def on_note_arrange(self, *args):
    if not self.canvas: return
    class ArrangeAtFretOptions(object):
      def __init__(self):
        self.fret          = 0
        self.details       = _("Arrange selected notes at fret?")
        self.icon_filename = os.path.join(globdef.DATADIR, "song.png")
      def __str__(self): return _("Arrange notes at fret...")
    
    o = ArrangeAtFretOptions()
    
    def on_validate(o):
      if o: self.canvas.arrange_selection_at_fret(o.fret)
        
    editobj3.edit(o, on_validate = on_validate, menubar = False)
        
  def on_dump(self, *args):
    if not self.song: return
    print(self.song.__xml__().getvalue(), file = sys.stderr)
    
  def on_gc(self, *args):
    import gc
    print(gc.collect(), "object(s) collected", file = sys.stderr)
    for o in gc.garbage: print("Garbage:", o, file = sys.stderr)
    print(file = sys.stderr)
    
    #import memtrack, weakref
    #ref = weakref.ref(self.songbook.song_refs[0].get_song())
    #if ref(): memtrack.reverse_track(ref())
    
    scan()
    
  def set_duration_longer (self): self.set_duration_longer_or_shorter(True)
  def set_duration_shorter(self): self.set_duration_longer_or_shorter(False)
  def set_duration_longer_or_shorter(self, longer):
    if not self.canvas: return
    if longer: notes = [note for note in self.canvas.selections if note.base_duration() < 384]
    else:      notes = [note for note in self.canvas.selections if note.base_duration() > 12 ]
    def decrease_duration(notes = notes):
      for note in notes: self.canvas.current_duration = note.duration = note.duration // 2
    def increase_duration(notes = notes):
      for note in notes: self.canvas.current_duration = note.duration = note.duration *  2
      
    if longer: undoredo.UndoableOperation(increase_duration, decrease_duration, _("increase note duration"), self.undo_stack)
    else:      undoredo.UndoableOperation(decrease_duration, increase_duration, _("decrease note duration"), self.undo_stack)
    
  def set_note_menus_enabled(self, enabled):
    # Needed because Qt considers AltGr + 8 (= backslash \) as a "_" for menu accels
    for menu in self.fx_menus.values(): self.set_menu_enable(menu, enabled)
    
  def on_pipe_command_received(self, info_message):
    print("EMITTED", info_message)
    




class InstrumentChooserPane(qtwidgets.QWidget):
  def __init__(self, main):
    qtwidgets.QWidget.__init__(self)
    layout    = qtwidgets.QVBoxLayout()
    
    layout.addWidget(qtwidgets.QLabel(), 3)
    
    label     = qtwidgets.QLabel("%s\n " % _("Choose an instrument:"))
    layout.addWidget(label, 0, qtcore.Qt.AlignHCenter)
    
    image     = qtwidgets.QLabel()
    pixmap    = qtgui.QPixmap(os.path.join(globdef.DATADIR, "instruments.png"), "png")
    image.setPixmap(pixmap)
    image.mouseReleaseEvent = self.on_mouse_release
    layout.addWidget(image, 0, qtcore.Qt.AlignHCenter)
    
    layout.addWidget(qtwidgets.QLabel(), 2)
    
    self.setLayout(layout)
    self.main         = main
    self.image        = image
    self.pixmap_width = pixmap.width()
    
  def on_mouse_release(self, event):
    x = event.x()
    if not(0 <= x < self.pixmap_width): return
    
    s = new_song()
    p = model.Partition(s)
    s.partitions.append(p)
    if   x < 100:
      p.view = model.GuitarView(p)
      sound_file = "son_guitare.mid"
    elif x < 168:
      from songwrite3.plugins.lyre import SevenStringLyreView
      p.view = SevenStringLyreView(p)
      sound_file = "son_lyre.mid"
    elif x < 255:
      p.view = model.VocalsView(p)
      sound_file = "son_chant.mid"
    elif x < 362:
      from songwrite3.plugins.accordion import AccordionView
      from songwrite3.plugins.chord     import AccordionChordView
      p.view = AccordionView(p)
      p2 = model.Partition(s)
      s.partitions.append(p2)
      p2.view = AccordionChordView(p2)
      p2.volume = 128
      sound_file = "son_accordeon.mid"
    else:
      from songwrite3.plugins.fingering import TinWhistleView
      p.view = TinWhistleView(p)
      sound_file = "son_flute.mid"
      
    player.play(open(os.path.join(globdef.DATADIR, sound_file), "rb").read())
    
    for p in s.partitions: p.instrument = p.view.default_instrument
    self.main.filename = None
    self.main.set_song(s)
    
    
def new_song():
  s = model.Song()
  s.authors   = getpass.getuser().title()
  s.title     = _("%s's song") % s.authors
  s.copyright = "Copyright %s by %s" % (time.localtime(time.time())[0], s.authors)
  return s



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2008 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import editobj3, editobj3.introsp as introsp, editobj3.field as field
import songwrite3.model   as model
import songwrite3.plugins as plugins

class FingeringView(model.View):
  view_class               = "fingering"
  default_instrument       = 72
  default_icon_filename    = "flute.png"
  
  def __init__(self, partition, name = ""):
    model.View.__init__(self, partition, name)
    if not hasattr(partition, "print_with_staff_too"): partition.print_with_staff_too = 0
    
  def get_drawer(self, canvas, compact = False):
    from songwrite3.plugins.fingering.drawer import FingeringDrawer, SingleString
    return FingeringDrawer(canvas, self.partition, compact, SingleString)

  def get_type(self):
    return self.__class__
  
  def note_2_fingering(self, note):
    v = abs(note.value)
    if getattr(self.partition, "g8", 0): v += 12
    return self.fingerings.get(v)
  

class TinWhistleView(FingeringView):
  default_instrument    = 72
  default_icon_filename = "flute_irlandaise.png"
  fingerings = {
    #note : (hole1, hole2, hole3, hole4, hole5, hole6,   breath),
    62 : (1  , 1  , 1  , None, 1  , 1  , 1  ,     ""),
    63 : (1  , 1  , 1  , None, 1  , 1  , 0.5,     ""),
    64 : (1  , 1  , 1  , None, 1  , 1  , 0  ,     ""),
    65 : (1  , 1  , 1  , None, 1  , 0.5, 0  ,     ""),
    66 : (1  , 1  , 1  , None, 1  , 0  , 0  ,     ""),
    67 : (1  , 1  , 1  , None, 0  , 0  , 0  ,     ""),
    68 : (1  , 1  , 0.5, None, 0  , 0  , 0  ,     ""),
    69 : (1  , 1  , 0  , None, 0  , 0  , 0  ,     ""),
    70 : (1  , 0.5, 0  , None, 0  , 0  , 0  ,     ""),
    71 : (1  , 0  , 0  , None, 0  , 0  , 0  ,     ""),
    72 : (0  , 1  , 1  , None, 1  , 0  , 1  ,     ""),
    73 : (0  , 0  , 0  , None, 0  , 0  , 0  ,     ""),
    
    74 : (0  , 1  , 1  , None, 1  , 1  , 1  ,     "+"),
    75 : (1  , 1  , 1  , None, 1  , 1  , 0.5,     "+"),
    76 : (1  , 1  , 1  , None, 1  , 1  , 0  ,     "+"),
    77 : (1  , 1  , 1  , None, 1  , 0.5, 0  ,     "+"),
    78 : (1  , 1  , 1  , None, 1  , 0  , 0  ,     "+"),
    79 : (1  , 1  , 1  , None, 0  , 0  , 0  ,     "+"),
    80 : (1  , 1  , 0.5, None, 0  , 0  , 0  ,     "+"),
    81 : (1  , 1  , 0  , None, 0  , 0  , 0  ,     "+"),
    82 : (1  , 0.5, 0  , None, 0  , 0  , 0  ,     "+"),
    83 : (1  , 0  , 0  , None, 0  , 0  , 0  ,     "+"),
    85 : (0  , 0  , 0  , None, 0  , 0  , 0  ,     "+"),
    }
  
  def __init__(self, partition, name = ""):
    if not hasattr(partition, "instrument_tonality"):
      partition.instrument_tonality = "D"
      if not partition.notes: partition.tonality = "D"
    FingeringView.__init__(self, partition, name or _("Tin whistle"))
    
  def get_icon_filename(self): return "flute_irlandaise.png"
  
  def note_2_fingering(self, note):
    v = abs(note.value)
    if getattr(self.partition, "g8", 0): v += 12
    return self.fingerings.get(v - model.OFFSETS[self.partition.instrument_tonality] + 2)
  
  def get_bell_note(self):
    bell_note = 62 + model.OFFSETS[self.partition.instrument_tonality] - 2
    if self.partition.g8: bell_note -= 12
    return bell_note
  
  def get_drawer(self, canvas, compact = False):
    from songwrite3.plugins.fingering.drawer import FingeringDrawer, TinWhistleSingleString
    return FingeringDrawer(canvas, self.partition, compact, TinWhistleSingleString)
  
  def __xml__(self, xml = None, context = None):
    xml.write(u'''\t\t<view type="tin_whistle"/>\n''')
    
    
class RecorderView(FingeringView):
  default_instrument = 73
  fingerings = {
    #note : (hole1, hole2, hole3, hole4, hole5, hole6,   breath),
    60 : (1  , None, 1  , 1  , 1  , None, 1  , 1  , (1, 1), (1, 1),    ""),
    61 : (1  , None, 1  , 1  , 1  , None, 1  , 1  , (1, 1), (1, 0),    ""),
    62 : (1  , None, 1  , 1  , 1  , None, 1  , 1  , (1, 1), (0, 0),    ""),
    63 : (1  , None, 1  , 1  , 1  , None, 1  , 1  , (1, 0), (0, 0),    ""),
    64 : (1  , None, 1  , 1  , 1  , None, 1  , 1  , (0, 0), (0, 0),    ""),
    65 : (1  , None, 1  , 1  , 1  , None, 1  , 0  , (1, 1), (1, 1),    ""),
    66 : (1  , None, 1  , 1  , 1  , None, 0  , 1  , (1, 1), (0, 0),    ""),
    67 : (1  , None, 1  , 1  , 1  , None, 0  , 0  , (0, 0), (0, 0),    ""),
    68 : (1  , None, 1  , 1  , 0  , None, 1  , 1  , (0, 0), (0, 0),    ""),
    69 : (1  , None, 1  , 1  , 0  , None, 0  , 0  , (0, 0), (0, 0),    ""),
    70 : (1  , None, 1  , 0  , 1  , None, 1  , 0  , (0, 0), (0, 0),    ""),
    71 : (1  , None, 1  , 0  , 0  , None, 0  , 0  , (0, 0), (0, 0),    ""),
    
    72 : (1  , None, 0  , 1  , 0  , None, 0  , 0  , (0, 0), (0, 0),    ""),
    73 : (0  , None, 1  , 1  , 0  , None, 0  , 0  , (0, 0), (0, 0),    ""),
    74 : (0  , None, 0  , 1  , 0  , None, 0  , 0  , (0, 0), (0, 0),    ""),
    75 : (0  , None, 1  , 1  , 1  , None, 1  , 1  , (1, 1), (1, 1),    ""),
    76 : (0.5, None, 1  , 1  , 1  , None, 1  , 1  , (0, 0), (0, 0),    ""),
    77 : (0.5, None, 1  , 1  , 1  , None, 1  , 0  , (1, 1), (0, 0),    ""),
    78 : (0.5, None, 1  , 1  , 1  , None, 0  , 1  , (0, 0), (0, 0),    ""),
    79 : (0.5, None, 1  , 1  , 1  , None, 0  , 0  , (0, 0), (0, 0),    ""),
    80 : (0.5, None, 1  , 1  , 0  , None, 1  , 0  , (0, 0), (0, 0),    ""),
    81 : (0.5, None, 1  , 1  , 0  , None, 0  , 0  , (0, 0), (0, 0),    ""),
    82 : (0.5, None, 1  , 1  , 0  , None, 1  , 1  , (1, 1), (0, 0),    ""),
    83 : (0.5, None, 1  , 1  , 0  , None, 1  , 1  , (0, 0), (0, 0),    ""),
    
    84 : (0.5, None, 1  , 0  , 0  , None, 1  , 1  , (0, 0), (0, 0),    ""),
    85 : (0.5, None, 1  , 0  , 1  , None, 1  , 0  , (0, 0), (0, 0),    ""),
    86 : (0.5, None, 1  , 0  , 1  , None, 1  , 0  , (1, 1), (0, 0),    ""),
    87 : (0.5, None, 0  , 1  , 1  , None, 0  , 1  , (1, 1), (0, 0),    ""),
    }
  
  def __init__(self, partition, name = ""):
    FingeringView.__init__(self, partition, name or _("Recorder"))
    
  def get_icon_filename(self): return "flute.png"
  
  def get_bell_note(self):
    bell_note = 60
    if self.partition.g8: bell_note -= 12
    return bell_note
  
  
  def __xml__(self, xml = None, context = None):
    xml.write(u'''\t\t<view type="recorder"/>''')
    
  
model.VIEW_CATEGORIES.append("fingering")
plugins.ViewPlugin(TinWhistleView, None, "tin_whistle", "fingering")
plugins.ViewPlugin(RecorderView  , None, "recorder"   , "fingering")

#descr = introsp.description(model.Partition)
#descr.set_field_for_attr("instrument_tonality"  , field.EnumField(dict([(_("tonality_%s" % tonality), tonality) for tonality in model.TONALITIES.keys()])))





# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2007-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, math, bisect, locale, codecs
from io import StringIO

import editobj3.editor_qt as editor_qt

import songwrite3.model  as model
from   songwrite3.canvas import *

import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtGui     as qtgui


class PDFAdditionalStaffDrawer(StaffDrawer):
  def draw(self, ctx, x, y, width, height, drag = 0):
    self.partition.real_view = self.partition.view
    try:
      self.partition.view = model.GuitarStaffView(self.partition, 0)
      
      StaffDrawer.draw(self, ctx, x, y, width, height, drag)
    finally:
      self.partition.view = self.partition.real_view
      del self.partition.real_view
      
  def draw_note(self, ctx, note, string_id, y):
    if (note.fx == "harmonic") and (self.partition.real_view.use_harmonics_for_octavo): # Lyre use harmonic for second octavo, not as fx
      note.fx = ""
      StaffDrawer.draw_note(self, ctx, note, string_id, y)
      note.fx = "harmonic"
    else:
      StaffDrawer.draw_note(self, ctx, note, string_id, y)
    
    
class PDFLyricsDrawer(Drawer):
  def __init__(self, canvas, lyrics, melody):
    Drawer.__init__(self, canvas)
    
    self.lyrics  = lyrics
    self.melody  = melody
    self.x0      = canvas.start_x + 2
    self.start_y = 0
    
  def draw(self, ctx, x, y, width, height, drag = 0):
    if Drawer.draw(self, ctx, x, y, width, height, drag):
      text_y = self.y + self.start_y + 1 + self.canvas.default_line_height / 2 + self.canvas.default_ascent / 2 - self.canvas.default_descent / 2
      lignes = self.lyrics.text.split("\n")
      
      nb_ligne = 1
      ctx.setPen(qtcore.Qt.black)
      cur_x = 0
      
      reduced_fonts = []
      font_size = self.canvas.default_font_size
      while font_size > 3:
        font = qtgui.QFont("Sans", font_size)
        font = qtgui.QFont(font, ctx.device())
        font.setPixelSize(font_size)
        metrics = qtgui.QFontMetrics(font)
        reduced_fonts.append((font, metrics))
        font_size -= 1
        
      def render_syllabe(syllabe, text_x, text_y, max_width):
        for reduced_font, reduced_metrics in reduced_fonts:
          width = reduced_metrics.width(syllabe)
          if width < max_width: break
        ctx.setFont(reduced_font)
        ctx.drawText(text_x, text_y, syllabe)
        
      syllabe = ""
      for ligne in lignes:
        syllabes = ligne.replace("\\\\", "").replace("--", "-").split("\t")
        
        melody_notes0 = [note for note in self.melody.notes if not note.duration_fx == "appoggiatura"]
        melody_notes  = []
        previous_time = -1000
        for note in melody_notes0:
          if note.time == previous_time: continue
          melody_notes.append(note)
          previous_time = note.time
          
        nb_skip = 0
        for i in range(min(len(melody_notes), len(syllabes))):
          if nb_skip:
            nb_skip -= 1
            continue
          note    = melody_notes[i]
          if not (self.canvas.time1 <= note.time < self.canvas.time2):
            continue
          
          if syllabe:
            if syllabes[i] in ("_", "_-"):
              max_text_width += note.duration * self.canvas.zoom
              continue
            else:
              render_syllabe(syllabe, text_x, text_y, max_text_width)
              
          syllabe = syllabes[i]
          if not syllabe: continue
          
          text_x = self.canvas.time_2_x(note.time) + 1
          if text_x < x: continue
          if text_x > x + width: break
          
          if text_x <= cur_x: # Need breakline !
            nb_ligne += 1
            text_y   += self.canvas.default_line_height
          max_text_width = note.duration * self.canvas.zoom - 3 * self.scale
          
          cur_x = text_x
          
      if syllabe: render_syllabe(syllabe, text_x, text_y, max_text_width)
      
      self.height = self.start_y + nb_ligne * self.canvas.default_line_height + 1
      ctx.setFont(self.canvas.default_font)

    
class PDFCanvas(BaseCanvas):
  is_gui_interface = 0
  def __init__(self, song):
    BaseCanvas.__init__(self, song)
    
    self.set_default_font_size(song.printfontsize)
    self.update_mesure_size()
    
  def render(self, filename, x, width, time1, time2, zoom):
    self.filename = filename
    self.time1    = time1
    self.time2    = time2
    self.zoom     = zoom
    self.x        = x
    self.width    = width
    self.height   = 10000
    
    self.drawers  = []
    
    for i in range(2): # First and second pass for computing size
      buffer = qtcore.QBuffer()
      buffer.open(qtcore.QIODevice.WriteOnly)
      self.draw(buffer)
    self.draw(self.filename)
    
    
  def set_default_font_size(self, size, device = None):
    
    BaseCanvas.set_default_font_size(self, size, device or self.create_pdf_device(qtcore.QBuffer(), 10, 10))
  #  self.start_x = int((editor_qt.SMALL_ICON_SIZE + 20.0) * self.scale)
    
  def render_all(self): pass
  
  def create_pdf_device(self, filename, width, height):
    surface = qtgui.QPdfWriter(filename)
    
    if hasattr(qtgui, "QPageLayout"): # Qt >= 5.3
      surface.setResolution(72)
      page_layout = qtgui.QPageLayout()
      page_layout.setMode(qtgui.QPageLayout.FullPageMode)
      page_layout.setMargins(qtcore.QMarginsF(0.0, 0.0, 0.0, 0.0))
      page_layout.setPageSize(qtgui.QPageSize(qtcore.QSize(min(width, height), max(width, height)), None, qtgui.QPageSize.ExactMatch))
      if width > height: page_layout.setOrientation(qtgui.QPageLayout.Landscape)
      else:              page_layout.setOrientation(qtgui.QPageLayout.Portrait)
      surface.setPageLayout(page_layout)
      
    else: # Qt <= 5.2 
      surface.setMargins(qtgui.QPagedPaintDevice.Margins())
      #surface.setPageSizeMM(qtcore.QSizeF(width / 72 * 25.4, height / 72 * 25.4))
      surface.setPageSizeMM(qtcore.QSizeF(width / 1200.0 * 25.4, height / 1200.0 * 25.4))
      
    return surface
  
  def draw(self, filename):
    drawer = None

    surface = self.create_pdf_device(filename, self.width, self.height)
    self.set_default_font_size(self.song.printfontsize, surface)
    
    ctx = qtgui.QPainter(surface)
    ctx.setRenderHints(qtgui.QPainter.Antialiasing | qtgui.QPainter.TextAntialiasing | qtgui.QPainter.SmoothPixmapTransform)
    ctx.setBrush(qtcore.Qt.black)
    ctx.setFont(self.default_font)
    
    lyricss = []
    def add_lyrics():
      text = "\n".join([lyrics.text for lyrics in lyricss if lyrics.show_all_lines_on_melody])
      if text:
        cumulated_lyrics = model.Lyrics(self.song, text, _("Lyrics"))
        drawer = PDFLyricsDrawer(self, cumulated_lyrics, melody)
        self.drawers.append(drawer)
        
      else:
        syllabess = [line.split("\t") for lyrics in lyricss for line in lyrics.text.split("\n") if not lyrics.show_all_lines_on_melody]
        
        if syllabess:
          cummulated_syllabes = [""] * len([note for note in melody.notes if not note.duration_fx == "appoggiatura"])
          for i in range(len(cummulated_syllabes)):
            for syllabes in syllabess:
              if (i < len(syllabes)) and syllabes[i]:
                cummulated_syllabes[i] = syllabes[i]
                break
          cumulated_lyrics = model.Lyrics(self.song, "\t".join(cummulated_syllabes), _("Lyrics"))
          drawer = PDFLyricsDrawer(self, cumulated_lyrics, melody)
          self.drawers.append(drawer)
          
      lyricss.__imul__(0)
      
    if not self.drawers:
      for partition in self.song.partitions:
        if   isinstance(partition, model.Partition):
          add_lyrics()
          melody = partition
          if not partition.notes_at(self.time1, self.time2 - 1): continue
          
        if   getattr(partition, "print_with_staff_too", 0) and not isinstance(partition.view, model.StaffView):
          drawer = PDFAdditionalStaffDrawer(self, partition, compact = 1)
          self.drawers.append(drawer)
          
        if isinstance(partition, model.Lyrics): lyricss.append(partition)
        else: drawer = partition.view.get_drawer(self, compact = 1)
        if drawer:
          if   getattr(partition, "print_with_staff_too", 0): drawer.show_header = 0
          self.drawers.append(drawer)
          drawer = None
      add_lyrics()
      
      for drawer in self.drawers: drawer.drawers_changed()
      
    if not self.drawers:
      for partition in self.song.partitions:
        if   isinstance(partition, model.Partition):
          add_lyrics()
          melody = partition
          
        if   getattr(partition, "print_with_staff_too", 0) and not isinstance(partition.view, model.StaffView):
          drawer = PDFAdditionalStaffDrawer(self, partition, compact = 1)
          self.drawers.append(drawer)
          
        if isinstance(partition, model.Lyrics): lyricss.append(partition)
        else: drawer = partition.view.get_drawer(self, compact = 1)
        if drawer:
          if   getattr(partition, "print_with_staff_too", 0): drawer.show_header = 0
          self.drawers.append(drawer)
          drawer = None
        break # Only first
      
      add_lyrics()
      
      for drawer in self.drawers: drawer.drawers_changed()
      
    x     = self.start_x
    width = self.width
    y     = 0
    y1    = 0
    y2    = 0
    dy    = 0
    total_height = 0
    for drawer in self.drawers:
      drawer.y = dy
      drawer.draw(ctx, x, y, width, self.height)
      dy           += drawer.height
      total_height += drawer.height
      
    # Show mesure number at the beginning
    if self.drawers and isinstance(self.drawers[0], PartitionDrawer):
      mesure_number = str(1 + self.song.mesures.index(self.song.mesure_at(self.time1)))
      ctx.setFont(self.small_font)
      text_size = ctx.fontMetrics().size(0, mesure_number)
      ctx.drawText(
        self.start_x - text_size.width() // 2,
        self.drawers[0].y + self.drawers[0].start_y + self.drawers[0].stem_extra_height - text_size.height() // 2,
        mesure_number)
      
    for drawer in self.drawers:
      if hasattr(drawer, "strings") and (len(drawer.strings) > 1):
        y1 = drawer.y + drawer.start_y + drawer.stem_extra_height + drawer.string_height * 1.5
        break
    for drawer in reversed(self.drawers):
      if hasattr(drawer, "strings") and (len(drawer.strings) > 1):
        y2 = drawer.y + drawer.start_y + drawer.stem_extra_height + drawer.string_height * 1.5
        break
      
    mesure = self.song.mesure_at(self.time1)
    x = self.start_x
    #x = self.time_2_x(self.time1)
    #x = self.start_x - self.x + self.time1 * self.zoom
    ctx.drawLine(x, y1, x, y2)
    
    ctx.end()
    self.height = total_height



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2011 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from editobj3.undoredo import *

import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtGui     as qtgui

import songwrite3.model     as model
import songwrite3.globdef   as globdef
import songwrite3.canvas    as canvas
from songwrite3.plugins.chord.chord import ChordManager, TablatureChordManager, LyreChordManager, PianoChordManager, AccordionChordManager

def set_chord(canvas, partition, chord_manager, times, base, second = 0, third = 1, fourth = 0, fifth = 7, sixth = 0, seventh = 0):
  if third == 1: third = chord_manager.choose_third(partition, base)
    #if ((base + 4 - model.OFFSETS[partition.tonality]) % 12) in model.NOTES: third = 4
    #else:                                                                    third = 3
  if not isinstance(times, dict): # times should be a dict mapping each time to the reference note for it
    times2 = {}
    for time in times:
      notes = partition.notes_at(time)
      times2[time] = notes[0]
    times = times2
  values, frets  = chord_manager.create(base, second, third, fourth, fifth, sixth, seventh)
  saved          = {}
  saved_duration = {}
  removed        = []
  new_notes      = []
  def do_it(saved_duration = saved_duration):
    for time, ref in times.items():
      notes = partition.notes_at(time)
      i = 0
      chord_notes = []
      for value, string_id in values:
        if i < len(notes):
          saved[notes[i]] = notes[i].value, getattr(notes[i], "string_id", 0)
          notes[i].value = value
          if string_id != -1:
            notes[i].string_id = string_id
          chord_notes.append(notes[i])
          if canvas:
            canvas.auto_update_duration(notes[i], saved_duration)
            canvas.auto_update_duration(notes[i].partition.note_after(notes[i]), saved_duration)
          i += 1
        else:
          new_note = model.Note(partition, time, ref.duration, value, ref.volume)
          if string_id != -1:
            new_note.string_id  = string_id
          new_note.fx           = ref.fx
          new_note.link_fx      = ref.link_fx
          new_note.duration_fx  = ref.duration_fx
          new_note.strum_dir_fx = ref.strum_dir_fx
          new_notes  .append(new_note)
          chord_notes.append(new_note)
      if len(values) < len(notes):
        removed.extend(notes[len(values) - len(notes) :])
    partition.remove(*removed)
    partition.add_note(*new_notes)
    if canvas:
      for new_note in new_notes:
        canvas.auto_update_duration(new_note, saved_duration)
        canvas.auto_update_duration(new_note.partition.note_after(new_note), saved_duration)
      canvas.deselect_all()
      for note in chord_notes: canvas.partition_2_drawer[partition].select_note(note)
      
  def undo_it(saved_duration = saved_duration):
    for note, (value, string_id) in saved.items():
      note.value     = value
      note.string_id = string_id
    if removed  : partition.add_note(*removed)
    if new_notes: partition.remove  (*new_notes)
    if canvas: canvas.restore_saved_duration(saved_duration)
    saved         .clear()
    saved_duration.clear()
    removed       .__imul__(0)
    new_notes     .__imul__(0)
  return do_it, undo_it


class ChordDrawer(canvas.PartitionDrawer):
  avoid_repeat      = 1
  always_draw_stems = 1
  need_extra_space_for_strum_dir = 0
  def __init__(self, canvas_, partition, compact = False):
    canvas.PartitionDrawer.__init__(self, canvas_, partition, compact)
    self.strings = [SingleString(self)]
    self.string_height     = self.canvas.default_line_height * 1.3
    self.stem_extra_height = self.canvas.default_line_height
    self.chord_manager     = self.ChordManager(partition)
    self.last_chord_name   = None
    self.last_chord_mesure = None
    self.note_offset_x     = 3.0 * self.scale
    
  def config_listener(self, obj, type, new, old):
    canvas.PartitionDrawer.config_listener(self, obj, type, new, old)
    if type is object:
      if   (new.get("ENGLISH_CHORD_NAME") != old.get("ENGLISH_CHORD_NAME")):
        self.chord_manager = self.ChordManager(self.partition)
        self.canvas.render_all()
        
  def partition_listener(self, partition, type, new, old):
    canvas.PartitionDrawer.partition_listener(self, partition, type, new, old)
    if type is object:
      if   (new.get("tonality") != old.get("tonality")):
        self.chord_manager = self.ChordManager(self.partition)
        self.canvas.render_all()
        
  def on_touchscreen_new_note(self, event): return False
  
  def note_width(self, note): return 9.0 * self.scale
  
  def select_note(self, note):
    notes = self.partition.notes_at(note)
    if not note in notes:
      self.canvas.add_selection(note, *self.note_dimensions(note))
    for note in notes:
      self.canvas.add_selection(note, *self.note_dimensions(note))
      
  def on_key_press(self, event):
    keyval = event.key()
    if   qtcore.Qt.Key_A <= keyval <= qtcore.Qt.Key_G: # a-g
      value = model.NOTES[(keyval - qtcore.Qt.Key_A + 5) % 7]
      if value in model.TONALITIES[self.partition.tonality]:
        if model.TONALITIES[self.partition.tonality][0] == model.DIESES[0]:
          value += 1
        else:
          value -= 1
          
      times = {}
      for note in self.canvas.selections: times[note.time] = note
      do_it, undo_it = set_chord(self.canvas, self.partition, self.chord_manager, times, value)
      UndoableOperation(do_it, undo_it, _("add chord"), self.canvas.main.undo_stack)
      return 1
    
    elif keyval == qtcore.Qt.Key_M: # m
      self.toggle_major()
      return 1
    
    elif qtcore.Qt.Key_0 <= keyval <= qtcore.Qt.Key_9: # Numbers
      nb = keyval - qtcore.Qt.Key_0
      if nb == 5: self.toggle_5()
      if nb == 7: self.toggle_7()
      return 1
    
  def toggle_major(self):
    times = {}
    for note in self.canvas.selections: times[note.time] = 1
    for time in times:
      notes      = self.partition.notes_at(time)
      rel_values = frozenset([note.value for note in notes])
      chord      = self.chord_manager.identify(rel_values)
      if chord:
        third = chord.identified_notes[2]
        if third == 3: third = 4; break
        else:          third = 3; break
    else: third = 4
    do_its   = []
    undo_its = []
    for time in times:
      notes      = self.partition.notes_at(time)
      rel_values = frozenset([note.value for note in notes])
      chord      = self.chord_manager.identify(rel_values)
      v = [value or 0 for value in chord.identified_notes]
      #do_it1, undo_it1 = self.set_chord([time], chord.base, v[1], third, v[3], v[4], v[5], v[6], return_func = 1)
      do_it1, undo_it1 = set_chord(self.canvas, self.partition, self.chord_manager, [time], chord.base, v[1], third, v[3], v[4], v[5], v[6])
      do_its  .append(  do_it1)
      undo_its.append(undo_it1)
    def do_it():
      for func in do_its: func()
    def undo_it():
      for func in undo_its: func()
    UndoableOperation(do_it, undo_it, _("add chord"), self.canvas.main.undo_stack)
    
  def toggle_5(self):
    times = {}
    for note in self.canvas.selections: times[note.time] = 1
    for time in times:
      notes      = self.partition.notes_at(time)
      rel_values = frozenset([note.value for note in notes])
      chord      = self.chord_manager.identify(rel_values)
      if chord:
        third = chord.identified_notes[2]
        if third: third = 0; break
    else: third = 1
    
    do_its   = []
    undo_its = []
    for time in times:
      notes      = self.partition.notes_at(time)
      rel_values = frozenset([note.value for note in notes])
      chord      = self.chord_manager.identify(rel_values)
      v = [value or 0 for value in chord.identified_notes]
      do_it1, undo_it1 = set_chord(self.canvas, self.partition, self.chord_manager, [time], chord.base, v[1], third, v[3], v[4], v[5], v[6])
      do_its  .append(  do_it1)
      undo_its.append(undo_it1)
    def do_it():
      for func in do_its: func()
    def undo_it():
      for func in undo_its: func()
    UndoableOperation(do_it, undo_it, _("add chord"), self.canvas.main.undo_stack)
    
  def toggle_7(self):
    times = {}
    for note in self.canvas.selections: times[note.time] = 1
    for time in times:
      notes      = self.partition.notes_at(time)
      rel_values = frozenset([note.value for note in notes])
      chord      = self.chord_manager.identify(rel_values)
      if chord:
        seventh = chord.identified_notes[6]
        if seventh: seventh = 0; break
    else: seventh = 10
    do_its   = []
    undo_its = []
    for time in times:
      notes      = self.partition.notes_at(time)
      rel_values = frozenset([note.value for note in notes])
      chord      = self.chord_manager.identify(rel_values)
      v = [value or 0 for value in chord.identified_notes]
      #do_it1, undo_it1 = self.set_chord([time], chord.base, v[1], v[2], v[3], v[4], v[5], seventh, return_func = 1)
      do_it1, undo_it1 = set_chord(self.canvas, self.partition, self.chord_manager, [time], chord.base, v[1], v[2], v[3], v[4], v[5], seventh)
      do_its  .append(  do_it1)
      undo_its.append(undo_it1)
    def do_it():
      for func in do_its: func()
    def undo_it():
      for func in undo_its: func()
    UndoableOperation(do_it, undo_it, _("add chord"), self.canvas.main.undo_stack)
      
  # No "string" notion in chord notation!
  def note_string_id   (self, note ): return 0
  def default_string_id(self, value): return 0
  def y_2_string_id(self, y, force = 0):
    if force:
      if   self.y + self.start_y > y: return -1
      elif y > self.y + self.height:  return  1
      return 0
    else:
      if self.y + self.start_y <= y <= self.y + self.height: return 0
      return None
  def string_id_2_y(self, string_id = 0):
    return self.y + self.start_y + self.stem_extra_height + self.string_height // 2
  
  def draw(self, ctx, x, y, width, height, drag = 0, draw_icon = 1):
    if self.compact:
      canvas.PartitionDrawer.draw(self, ctx, x, y, width, height, drag, 0, -11.0)
    else:
      canvas.PartitionDrawer.draw(self, ctx, x, y, width, height, drag, 0, -11.0)
    
  def draw_mesure(self, ctx, time, mesure, y, height):
    canvas.PartitionDrawer.draw_mesure(self, ctx, time, mesure, y, height, False)
    #pass
  
  def draw_strings(self, ctx, x, y, width, height):
    self.height = int(self.start_y + self.stem_extra_height + self.string_height + self.stem_extra_height_bottom + 1)
    self.last_chord_name = None
    
  def render_note(self, note):
    mesure = self.canvas.song.mesure_at(note)
    if mesure.rythm2 == 8: time1 = (note.link_start() // 144) * 144; time2 = max(note.link_end(), time1 + 144)
    else:                  time1 = (note.link_start() //  96) *  96; time2 = max(note.link_end(), time1 +  96)
    self.canvas.render_pixel(
      self.canvas.time_2_x(time1) - 10.0 * self.scale,
      self.y,
      (time2 - time1) * self.canvas.zoom + 50.0 * self.scale,
      self.height + self.canvas.default_line_height // 2 + self.canvas.default_ascent // 2,
      )
    
  def note_text(self, note): return "_"
  
  def draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure = None, previous_time = -32000, next_time = 32000):
    x = x + self.canvas.default_line_height // 2
    canvas.PartitionDrawer.draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure, previous_time, next_time)
    
  def draw_stem_and_beam_for_chord(self, ctx, notes, notes_y, previous, next, appo = 0):
    self.draw_chord(ctx, notes, notes_y, previous, next, appo)
      
  def draw_chord(self, ctx, notes, notes_y, previous, next, appo):
    ctx.setPen(self.note_color(notes[0]))
    rel_values = frozenset([note.value for note in notes])
    if rel_values == frozenset([0]):
      ctx.drawText(self.canvas.time_2_x(notes[0].time) + self.note_offset_x, self.y + self.start_y + 2 * self.stem_extra_height, "_")
    else:
      chord = self.chord_manager.identify(rel_values)
      if chord: name = chord.name
      else:     name = "?"
      mesure = self.partition.song.mesure_at(notes[0].time)
      if (name != self.last_chord_name) or (mesure != self.last_chord_mesure):
        max_width = 99999
        if next:
          next_notes = self.partition.notes_at(next.time)
          rel_values = frozenset([note.value for note in next_notes])
          if rel_values:
            chord = self.chord_manager.identify(rel_values)
            if chord and ((not self.avoid_repeat) or (chord.name != name)):
              max_width = (self.canvas.time_2_x(next.time) - self.canvas.time_2_x(notes[0].time)) - 3 * self.scale
        ctx.drawText(
          self.canvas.time_2_x(notes[0].time) + self.note_offset_x,
          self.y + self.start_y + 2 * self.stem_extra_height - self.canvas.default_ascent,
          max_width,
          999.0,
          0,
          name,
        )
        
        if self.avoid_repeat:
          self.last_chord_name   = name
          self.last_chord_mesure = mesure
    if not self.draw_strum_direction(ctx, notes):
      canvas.PartitionDrawer.draw_stem_and_beam_for_chord(self, ctx, notes, notes_y, previous, next, appo)
    ctx.setPen(qtcore.Qt.black)
    
  def draw_strum_direction(self, ctx, notes, offset_x = 0.0):
    return canvas.PartitionDrawer.draw_strum_direction(self, ctx, notes, offset_x + self.note_offset_x)
    
  def draw_note(self, ctx, note, string_id, y): pass
  

class SingleString(object):
  def __init__(self, drawer, notation_pos = -1):
    self.drawer       = drawer
    self.notation_pos = notation_pos
    
  base_note = 0
  
  def text_2_value(self, note, text): return self.base_note
  def value_2_text(self, note): return "/"
  def __str__(self): return "Single string"
  def width(self): return -1
  def on_click_at(self, hole): return "0"
        
    
class TablatureChordDrawer(ChordDrawer):
  ChordManager = TablatureChordManager
  
  
class PianoChordDrawer(ChordDrawer):
  ChordManager = PianoChordManager
  
  
class LyreChordDrawer(ChordDrawer):
  ChordManager = LyreChordManager
  
  
class AccordionChordDrawer(ChordDrawer):
  avoid_repeat = 0
  ChordManager = AccordionChordManager
  
  def __init__(self, canvas_, partition, compact = False):
    ChordDrawer.__init__(self, canvas_, partition, compact)
    self.note_offset_x = 3.0 * self.scale
    
  def drawers_changed(self):
    try:    previous = self.canvas.drawers[self.canvas.drawers.index(self) - 1]
    except: self.left_hand_drawer = None
    else:
      if previous.__class__.__name__ == "AccordionDrawer": self.left_hand_drawer = previous
      else:                                                self.left_hand_drawer = None
    if self.left_hand_drawer: self.show_header = 0
    else:                     self.show_header = 1
    
  def draw(self, ctx, x, y, width, height, drag = 0, draw_icon = 1):
    if self.show_header:
      return canvas.PartitionDrawer.draw(self, ctx, x, y, width, height, drag, 0, -34)
    else:
      #text_width = ctx.text_extents(_("LG"))[2]
      #ctx.move_to(85.0 * self.scale - text_width, self.y + 14.5 * self.scale)
      #ctx.show_text(_("LG"))
      if self.compact:
        ctx.drawText(65.0 * self.scale, self.y + 20.0 * self.scale, _("LG"))
        return canvas.PartitionDrawer.draw(self, ctx, x, y, width, height, drag, 0, -10.0)
      else:
        ctx.drawText(65.0 * self.scale, self.y + 20.0 * self.scale, _("LG"))
        return canvas.PartitionDrawer.draw(self, ctx, x, y, width, height, drag, 0, -20.0)
      
  def draw_stem_and_beam_for_chord(self, ctx, notes, notes_y, previous, next, appo = 0):
    self.draw_chord(ctx, notes, notes_y, previous, next, appo)
    
  def draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure = None, previous_time = -32000, next_time = 32000):
    pass
    
  def on_key_press(self, event):
    keyval = event.key()
    if (event.modifiers() == qtcore.Qt.ShiftModifier) and (qtcore.Qt.Key_A <= keyval <= qtcore.Qt.Key_G): # A-G
      value = model.NOTES[(keyval - qtcore.Qt.Key_A + 5) % 7]
      if value in model.TONALITIES[self.partition.tonality]:
        if model.TONALITIES[self.partition.tonality][0] == model.DIESES[0]:
          value += 1
        else:
          value -= 1
          
      times = {}
      for note in self.canvas.selections: times[note.time] = note
      do_it, undo_it = set_chord(self.canvas, self.partition, self.chord_manager, times, value, second = 0, third = 0, fourth = 0, fifth = 0, sixth = 0, seventh = 0)
      UndoableOperation(do_it, undo_it, _("add bass"), self.canvas.main.undo_stack)
      return 1
    
    return ChordDrawer.on_key_press(self, event)
    



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2016-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, atexit

try:
  import getpass
  username = getpass.getuser()
  PIPE = "/tmp/songwrite3_pipe_%s" % username
except:
  PIPE = ""

def is_already_running():
  if (not PIPE) or (not os.path.exists(PIPE)): return False
  return True
  
def create():
  global PIPE
  
  if PIPE:
    try:
      os.mkfifo(PIPE)
    except:
      print("WARNING: Cannot open pipe %s." % PIPE, file = sys.stderr)
    atexit.register(destroy)
    
def destroy():
  if PIPE: os.unlink(PIPE)


THREAD = None
def listen():
  global THREAD # QThread must be kept in memory
  
  import PyQt5.QtCore as qtcore, songwrite3.model as model
  
  class ListeningThread(qtcore.QThread):
    commandReceived = qtcore.pyqtSignal([str])
    def run(self):
      while 1:
        f = open(PIPE, "r")
        command = f.readline()[:-1]
        f.close()
        if command:
          print("Received from pipe:", command, file = sys.stderr)
          self.commandReceived.emit(command)
          
  THREAD = ListeningThread()
  THREAD.commandReceived.connect(on_pipe_command_received)
  THREAD.start()
  
def send(command):
  import threading
  
  class SendingThread(threading.Thread):
    def __init__(self):
      threading.Thread.__init__(self)
      self.done = False
      
    def run(self):
      f = open(PIPE, "w")
      if self.done: return # Timeouted
      f.write(command + "\n")
      f.close()
      self.done = True
      
  thread = SendingThread()
  thread.start()
  thread.join(0.5) # Time out
  if not thread.done:
    thread.done = True # Prevent the thread from writing later
    return False
  else: return True
  
def on_pipe_command_received(command):
  try:
    if   command.startswith("open "):
      import songwrite3.model as model, songwrite3.main as main
      song = model.get_song(command[5:])
      main.App(song)
    elif command.startswith("new "):
      import songwrite3.main as main
      main.App()
    else:
      print("WARNING: Unknown command received!")
      
  except: sys.excepthook(*sys.exc_info())
    
if __name__ == "__main__":
  send("open /home/jiba/audio/tablatures/cours_tristan/tristan_reve.sw.xml")



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2011 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import editobj3, editobj3.introsp as introsp, editobj3.field as field
import songwrite3.model   as model
import songwrite3.plugins as plugins

#model.Note.accompaniment = None
model.Note.button_rank   = None
model.Note.bellows_dir   = None # -1 Push, 1 draw
model.NOTE_ATTRS.append("button_rank")
model.NOTE_ATTRS.append("bellows_dir")

accordion_tonalities = {
  -5 : _("C/F"),
   0 : _("G/C"),
   2 : _("A/D"),
  }

class AccordionView(model.View):
  view_class               = "accordion"
  default_icon_filename    = "accordion.png"
  default_instrument       = 21
  link_type                = model.LINK_NOTES_DEFAULT
  can_paste_note_by_string = 1
  automatic_string_id      = 1
  def __init__(self, partition, name = ""):
    model.View.__init__(self, partition, name or _("Accordion"))
    if partition:
      if not hasattr(partition, "print_with_staff_too"): partition.print_with_staff_too = 0
      if not hasattr(partition, "accordion_tonalities"): partition.accordion_tonalities = 0
    self.strings = [
      AccordionLeftHandString (self, -1, { "1" : 51, "2" : 50, "3" : 55, "4" : 59, "5" : 62, "6" : 67, "7" : 71, "8" : 74, "9" : 79, "10" : 83, "11" : 86,
                                           "1'": 56, "2'": 55, "3'": 60, "4'": 64, "5'": 67, "6'": 72, "7'": 76, "8'": 79, "9'": 84, "10'": 88 }), # Poussé
      AccordionLeftHandString (self,  1, { "1" : 49, "2" : 54, "3" : 57, "4" : 60, "5" : 64, "6" : 66, "7" : 69, "8" : 72, "9" : 76, "10" : 78, "11" : 81,
                                           "1'": 48, "2'": 59, "3'": 62, "4'": 65, "5'": 69, "6'": 71, "7'": 74, "8'": 77, "9'": 81, "10'": 83}), # Tiré
      ]
  
  def note_string_id(self, note):
    if   note.bellows_dir == -1: return 0
    elif note.bellows_dir ==  1: return 1
    return 1
  
  def get_drawer(self, canvas, compact = False):
    from songwrite3.plugins.accordion.drawer import AccordionDrawer
    return AccordionDrawer(canvas, self.partition, compact)
  
  def get_icon_filename(self): return "accordion.png"
  
  def __xml__(self, xml = None, context = None):
    xml.write('''\t\t<view type="accordion">\n''')
    xml.write("""\t\t</view>\n""")
    

#class AccordionLeftHandString(object):
class AccordionLeftHandString(model.TablatureString):
  def __init__(self, view, bellows_dir, button2value):
    self.view          = view
    self.bellows_dir   = bellows_dir
    self.button2value  = button2value
    self.value2buttons = {}
    # First pass for button rank 0, then second pass for button rank 1
    for button, value in button2value.items():
      if not "'" in button:
        if value in self.value2buttons: self.value2buttons[value].append(button)
        else:                           self.value2buttons[value] = [button]
    for button, value in button2value.items():
      if "'" in button:
        if value in self.value2buttons: self.value2buttons[value].append(button)
        else:                           self.value2buttons[value] = [button]
    self.base_note = self.button2value["3"]
    
  def text_2_value(self, note, text):
    value = self.button2value.get(text)
    if value is None:
      if len(text) > 1:
        value = self.button2value.get(text[-1])
      if value is None: 
        value = self.button2value["3"]
    note.button_rank = text.count("'")
    note.bellows_dir = self.bellows_dir
    return value
  def value_2_text(self, note, change_string = 1):
    buttons = self.value2buttons.get(note.value)
    if buttons is None:
      if change_string:
        for string in self.view.strings:
          if string is self: continue
          text = string.value_2_text(note, change_string = 0)
          if text:
            note.bellows_dir = string.bellows_dir
            return text
      return "?"
    if len(buttons) == 1: return buttons[0]
    button_rank = note.button_rank
    if button_rank is None:
      if (note.partition.tonality == "G") or (note.partition.tonality == "D"): button_rank = 0
      else:                                                                    button_rank = 1
    return buttons[button_rank]
  
  def width(self): return 7
  
  #def __str__(self): return _(self.__class__.__name__) % model.note_label(self.base_note)

#descr = introsp.description(AccordionLeftHandString)
introsp.def_attr("bellows_dir", field.HiddenField)

plugins.ViewPlugin(AccordionView, None, "accordion", "tab")



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2011 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

#from editobj3.undoredo import *
import songwrite3.model   as model
import songwrite3.plugins as plugins
from songwrite3.model import TablatureView, GuitarView, GuitarDADGADView, BassView, TablatureString

class ChordView(model.View):
  view_class               = "chord"
  saved_chords             = None
  can_paste_note_by_string = 0
  ok_for_lyrics            = 0
  
  def __init__(self, partition, name = ""):
    super(ChordView, self).__init__(partition, name)
    
  def _wrap_view_type(self, view_type):
    class MyView(view_type):
      saved_chords = None
    MyView.__name__ = view_type.__name__
    self._last_returned_view_type = MyView
    return MyView
    
  def get_type(self): return self._wrap_view_type(super(ChordView, self).get_type())
  
  def change_to_view_type(self, view_type):
    if issubclass(view_type, ChordView) and (not self.__class__ is view_type):
      from songwrite3.plugins.chord.drawer import set_chord
      new_view = view_type(self.partition)
      
      if view_type.saved_chords: # Restore previous chord (UNDO / REDO)
        view_type.saved_chords[0]()
        
      else: # Convert chord from one chord manager to another
        old_chord_manager = self     .get_Drawer().ChordManager(self.partition)
        new_chord_manager = view_type.get_Drawer().ChordManager(self.partition)
        times    = {}
        do_its   = []
        undo_its = []
        for note in self.partition.notes: times[note.time] = 1
        for time in times:
          notes      = self.partition.notes_at(time)
          rel_values = frozenset([note.value for note in notes])
          chord      = old_chord_manager.identify(rel_values)
          v = [value or 0 for value in chord.identified_notes]
          #new_chord_manager.create(chord.base, v[1], v[2], v[3], v[4], v[5], v[6])
          self.partition.view = new_view # HACK : new_chord_manager may use the partition's view, e.g. for strings attribute !
          do_it1, undo_it1 = set_chord(None, self.partition, new_chord_manager, [time], chord.base, v[1], v[2], v[3], v[4], v[5], v[6])
          self.partition.view = self # HACK : restore view
          do_its  .append(  do_it1)
          undo_its.append(undo_it1)
        previous_instrument = self.partition.instrument
        def do_it():
          self.partition.instrument = view_type.default_instrument
          for func in do_its: func()
        def undo_it():
          self.partition.instrument = previous_instrument
          for func in undo_its: func()
        do_it()
        self._last_returned_view_type.saved_chords = [undo_it]
      return new_view
    
  def get_drawer(self, canvas, compact = False): return self.get_Drawer()(canvas, self.partition, compact)
  
  def note_string_id(self, note): return 0
    
  
  
class TablatureChordView(ChordView, TablatureView):
  link_type = model.LINK_NOTES_DEFAULT
  @classmethod
  def get_Drawer(Class):
    from songwrite3.plugins.chord.drawer import TablatureChordDrawer
    return TablatureChordDrawer
    
  def __xml__(self, xml = None, context = None, type = "tablature_chord"):
    TablatureView.__xml__(self, xml, context, type)
    
class GuitarChordView(TablatureChordView, GuitarView):
  link_type = model.LINK_NOTES_DEFAULT
  default_instrument = 24
  default_icon_filename = "guitar.png"
  def __init__(self, partition, name = ""):
    super(GuitarChordView, self).__init__(partition, name)

from songwrite3.plugins.lyre import *
class LyreChordView(TablatureChordView):
  link_type = model.LINK_NOTES_DEFAULT
  default_instrument = 46
  default_icon_filename = "lyre.png"
  @classmethod
  def get_Drawer(Class):
    from songwrite3.plugins.chord.drawer import LyreChordDrawer
    return LyreChordDrawer
    
  def __xml__(self, xml = None, context = None, type = "lyre_chord"):
    TablatureView.__xml__(self, xml, context, type)
    
  def get_type(self):
    if   len(self.strings) == 6: return self._wrap_view_type(model.TablatureView.get_type(self, SixStringLyreChordView))
    else:                        return self._wrap_view_type(model.TablatureView.get_type(self, SevenStringLyreChordView))
    

class SixStringLyreChordView(LyreChordView, SixStringLyreView):
  link_type = model.LINK_NOTES_DEFAULT
  def __init__(self, partition, name = ""):
    super(SixStringLyreChordView, self).__init__(partition, name)

class SevenStringLyreChordView(LyreChordView, SevenStringLyreView):
  link_type = model.LINK_NOTES_DEFAULT
  def __init__(self, partition, name = ""):
    super(SevenStringLyreChordView, self).__init__(partition, name)


class PianoChordView(ChordView):
  default_instrument = 0
  default_icon_filename = "piano.png"
  def __init__(self, partition, name = ""):
    super(PianoChordView, self).__init__(partition, name or _("Piano"))
    
  @classmethod
  def get_Drawer(Class):
    from songwrite3.plugins.chord.drawer import PianoChordDrawer
    return PianoChordDrawer
    
  def __xml__(self, xml = None, context = None):
    xml.write(u'''\t\t<view type="piano_chord"''')
    xml.write(">\n")
    xml.write("""\t\t</view>\n""")
  

class AccordionChordView(ChordView):
  default_instrument = 21
  default_icon_filename = "accordion.png"
  def __init__(self, partition, name = ""):
    super(AccordionChordView, self).__init__(partition, name or _("Accordion"))
    
  @classmethod
  def get_Drawer(Class):
    from songwrite3.plugins.chord.drawer import AccordionChordDrawer
    return AccordionChordDrawer
    
  def __xml__(self, xml = None, context = None):
    xml.write(u'''\t\t<view type="accordion_chord"''')
    xml.write(">\n")
    xml.write("""\t\t</view>\n""")

model.VIEW_CATEGORIES.append("chords")
plugins.ViewPlugin(TablatureChordView      , TablatureString, "tablature_chord"        , None)
plugins.ViewPlugin(LyreChordView           , TablatureString, "lyre_chord"             , None)
plugins.ViewPlugin(GuitarChordView         , TablatureString, "guitar_chord"           , "chords")
plugins.ViewPlugin(SixStringLyreChordView  , TablatureString, "six_string_lyre_chord"  , "chords")
plugins.ViewPlugin(SevenStringLyreChordView, TablatureString, "seven_string_lyre_chord", "chords")
plugins.ViewPlugin(PianoChordView          , TablatureString, "piano_chord"            , "chords")
plugins.ViewPlugin(AccordionChordView      , TablatureString, "accordion_chord"        , "chords")





# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2007-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, os.path
from collections import OrderedDict
import editobj3, editobj3.introsp as introsp, editobj3.field as field
import songwrite3.globdef as globdef, songwrite3.model as model

INSTRUMENTS = [i.split(None, 1)[1] for i in _("__instru__").split("\n")]

INSTRUMENTS_2_ID = { instrument : i for (i, instrument) in enumerate(INSTRUMENTS) }

editobj3.TRANSLATOR = _


import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtGui     as qtgui

import editobj3.field_qt as field_qt, editobj3.editor_qt as editor_qt

editor_qt.SMALL_ICON_SIZE = 48



class QtDurationField(field_qt.QtField, qtwidgets.QWidget):
  def __init__(self, gui, master, obj, attr, undo_stack):
    qtwidgets.QWidget.__init__(self)
    super(QtDurationField, self).__init__(gui, master, obj, attr, undo_stack)
    self.options = []

    toolbar1 = qtwidgets.QToolBar()
    toolbar2 = qtwidgets.QToolBar()
    
    toolbar1.setIconSize(qtcore.QSize(48, 48))
    toolbar2.setIconSize(qtcore.QSize(48, 48))

    radio_group = qtwidgets.QActionGroup(self)
    for i in [1, 2, 3, 4, 5, 6]:
      if i < 3: toolbar = toolbar1
      else:     toolbar = toolbar2
      action = qtwidgets.QAction(toolbar)
      action.setIcon(qtgui.QIcon(os.path.join(globdef.DATADIR, "note_%s.png" % (6 * 2 ** (7 - i)))))
      action.setCheckable(True)
      radio_group.addAction(action)
      self.options.insert(0, action)
      toolbar.addAction(action)
    self.doted   = qtwidgets.QCheckBox(_("Dotted"))
    self.triplet = qtwidgets.QCheckBox(_("Triplet"))

    self.update()

    self.doted  .stateChanged.connect(self.validate)
    self.triplet.stateChanged.connect(self.validate)
    for option in self.options[::-1]: option.toggled.connect(self.validate)

    layout = qtwidgets.QVBoxLayout()
    layout.setContentsMargins(0, 0, 0, 0)
    layout.addWidget(toolbar1)
    layout.addWidget(toolbar2)
    layout2 = qtwidgets.QHBoxLayout()
    layout2.addWidget(self.doted)
    layout2.addWidget(self.triplet)
    layout.addLayout(layout2)
    self.setLayout(layout)

  def get_value(self):
    v = field.Field.get_value(self)
    if v is introsp.NonConsistent: return 0
    return v

  def validate(self, *args):
    for i, option in enumerate(self.options):
      if option.isChecked():
        v = 6 * 2 ** (i + 1)
        break
    else: v = 0
    if   self.doted  .isChecked(): v = int(v * 1.5)
    elif self.triplet.isChecked(): v = int(v / 1.5)
    self.set_value(v)

    if not self.updating:
      import songwrite3.main
      if isinstance(self.o, introsp.ObjectPack): partition = self.o.objects[0].partition
      else:                                      partition = self.o.partition
      songwrite3.main.APPS[partition.song].canvas.current_duration = v

  def update(self):
    self.updating += 1
    try:
      v = self.get_value()

      if v in (9, 18, 36, 72, 144, 288, 576):
        v = int(v / 1.5)
        self.doted.setChecked(True)
      else:
        self.doted.setChecked(False)

      if v in (4,  8, 16, 32,  64, 128, 256):
        v = int(v * 1.5)
        self.triplet.setChecked(True)
      else:
        self.triplet.setChecked(False)

      i = { 12 : 0, 24: 1, 48 : 2, 96 : 3, 192 : 4, 384 : 5 }.get(v)
      if not i is None: self.options[i].setChecked(1)
    finally: self.updating -= 1


DurationField  = QtDurationField

        
introsp.def_attr("details"      , field.HiddenField)
introsp.def_attr("icon_filename", field.HiddenField)

descr = introsp.description(globdef.Config)
descr.set_label  (_("Preferences"))
descr.set_details(_("__config_preamble__"))
descr.def_attr("MIDI_COMMAND"        , field.StringField)
descr.def_attr("PREVIEW_COMMAND_PDF" , field.StringField)
descr.def_attr("preview_command_pdf" , field.HiddenField)
descr.def_attr("PREVIEW_COMMAND_PDF2", field.HiddenField)
descr.def_attr("MIDI_USE_TEMP_FILE"  , field.HiddenField)
descr.def_attr("WINDOW_WIDTH"        , field.HiddenField)
descr.def_attr("WINDOW_HEIGHT"       , field.HiddenField)
descr.def_attr("WINDOW_X"            , field.HiddenField)
descr.def_attr("WINDOW_Y"            , field.HiddenField)

descr.def_attr("PAGE_FORMAT"         , field.EnumField({ _("A3") : "a3paper", _("A4") : "a4paper", _("A5") : "a5paper", _("letter") : "letterpaper", _("legal") : "legalpaper" }))
descr.def_attr("PAGE_MARGIN_TOP"     , field.FloatField)
descr.def_attr("PAGE_MARGIN_BOTTOM"  , field.FloatField)
descr.def_attr("PAGE_MARGIN_INTERIOR", field.FloatField)
descr.def_attr("PAGE_MARGIN_EXTERIOR", field.FloatField)

descr.def_attr("PREVIOUS_FILES"          , field.HiddenField)
descr.def_attr("ASK_FILENAME_ON_EXPORT"  , field.BoolField)
descr.def_attr("NUMBER_OF_PREVIOUS_FILES", field.IntField)
descr.def_attr("LAST_DIR"                , field.StringField)

#descr.def_attr("SIDE_BAR_POSITION", field.EnumField(["left", "right", "top", "bottom", "floating_horizontal", "floating_vertical", "none"]))
descr.def_attr("SIDE_BAR_POSITION", field.HiddenField)
descr.def_attr("AUTOMATIC_DURATION", field.BoolField)
descr.def_attr("PLAY_AS_TYPING", field.BoolField)
descr.def_attr("DISPLAY_PLAY_BAR", field.BoolField)
descr.def_attr("FONTSIZE", field.IntField)
descr.def_attr("ENGLISH_CHORD_NAME", field.BoolField)


def ask_new_song_ref(Class, songbook):
  import songwrite3.main as main
  filename = main.APPS[songbook].prompt_open_filename(".sw.xml")
  if not filename: return None
  song_ref = Class(songbook)
  song_ref.set_filename(filename)
  return song_ref

descr = introsp.description(model.Songbook)
descr.set_icon_filename(os.path.join(globdef.DATADIR, "songbook.png"))
descr.def_attr("song_refs", addable_values  = [introsp.NewInstanceOf(model.SongRef)], add_method = "insert_song_ref", remove_method = "remove_song_ref", label = "")
descr.def_attr("title"    , field.StringField)
descr.def_attr("authors"  , field.StringField)
descr.def_attr("comments" , field.TextField)
descr.def_attr("version"  , field.HiddenField)
descr.def_attr("filename" , field.HiddenField)

descr = introsp.description(model.SongRef)
descr.set_icon_filename(os.path.join(globdef.DATADIR, "song.png"))
descr.def_attr("filename", field.FilenameField)
descr.def_attr("title"   , field.HiddenField)
descr.def_attr("songbook", field.HiddenField)
descr.def_attr("song"    , field.HiddenField)
descr.set_constructor(introsp.Constructor(ask_new_song_ref))



def ask_new_partition(song):
  view_types = []
  for category in model.VIEW_CATEGORIES:
    for View in model.VIEWS[category]:
      view_types.append(introsp.NewInstanceOf(View, _(View.__name__), os.path.join(globdef.DATADIR, View.default_icon_filename)))
      
  return view_types
  

import songwrite3.plugins.pdf.pdf_latex

descr = introsp.description(model.Song)
descr.set_icon_filename(os.path.join(globdef.DATADIR, "song.png"))
descr.def_attr("partitions", addable_values = ask_new_partition, add_method = "insert_partition", remove_method = "remove_partition", label = "")
descr.def_attr("title"    , field.StringField)
descr.def_attr("authors"  , field.StringField)
descr.def_attr("copyright", field.StringField)
descr.def_attr("comments" , field.TextField)
descr.def_attr("filename" , field.HiddenField)
descr.def_attr("playlist" , field.HiddenField)
descr.def_attr("version"  , field.HiddenField)
descr.def_attr("mesures"  , field.HiddenField)
descr.def_attr("lang"     , field.EnumField(dict((_("__lang__%s" % code), code) for code in songwrite3.plugins.pdf.pdf_latex.lang_iso2latex.keys()), long_list = 0))
descr.def_attr("print_nb_mesures_per_line", field.RangeField(1, 20))
descr.def_attr("printfontsize", field.RangeField(6, 28))
descr.def_attr("print_with_staff_too", field.BoolField)
descr.def_attr("print_lyrics_columns", field.RangeField(1, 4))

descr = introsp.description(model.View)
descr.set_constructor(introsp.Constructor(lambda View, parent: model.Partition(parent, View.default_instrument, View)))

descr = introsp.description(model.ChorusView)
descr.set_constructor(introsp.Constructor(lambda View, parent: model.Lyrics(parent, header = "Chorus")))

def new_strophe(View, song):
  next_strophe_number = len([lyrics for lyrics in song.partitions if isinstance(lyrics, model.Lyrics) and _("Strophe #%s").replace("%s", "") in lyrics.header]) + 1
  return model.Lyrics(song, header = _("Strophe #%s") % next_strophe_number)
descr = introsp.description(model.StropheView)
descr.set_constructor(introsp.Constructor(new_strophe))

def note_2_base_note_names(note):
  if hasattr(note, "partition"): tonality = note.partition.tonality
  else:                          tonality = "C"
  d = OrderedDict()
  for i in range(12): d[model.note_label(i, False, tonality)] = i
  return d
  


def partition_2_icon_filename(self):
  return os.path.join(globdef.DATADIR, self.view.get_icon_filename())

VIEW_TYPE_NAMES = { _(View.__name__) : View for Views in model.VIEWS.values() for View in Views }

def view_value_2_enum(partition, view_type):
  if hasattr(view_type, "__name__"): return _(view_type.__name__)
  return ""

def view_enum_2_value(partition, view_type_name):
  return VIEW_TYPE_NAMES[view_type_name]

class Tuning(object):
  def __init__(self, strings):
    self.strings = strings
    self._string_class = strings[0].__class__

  def __str__(self): return _("Tuning")
  
  def insert(self, index, string): self.strings.insert(index, string)
    
  def remove(self, string): self.strings.remove(string)

def get_tuning(partition):
  if hasattr(partition.view, "strings"): return Tuning(partition.view.strings)
  return None

def new_tab_string(Class, tuning): return tuning._string_class()

descr = introsp.description(Tuning)
descr.set_icon_filename(os.path.join(globdef.DATADIR, "guitar.png"))
descr.def_attr("strings", addable_values = [introsp.NewInstanceOf(model.TablatureString)], add_method = "insert", remove = "remove", label = "")

def nb_sharp_or_bemol(tonality):
  alterations = model.TONALITIES[tonality]
  if not alterations: return ""
  if alterations[0] is model.DIESES[0]: return "\t(%s ♯)" % len(alterations)
  return "\t(%s ♭)" % len(alterations)

descr = introsp.description(model.Partition)
descr.set_icon_filename(partition_2_icon_filename)
descr.def_attr("header"    , field.TextField)
descr.def_attr("instrument", field.EnumField(INSTRUMENTS_2_ID))
descr.def_attr("tonality"           , field.EnumField({ _("tonality_%s" % tonality) + nb_sharp_or_bemol(tonality) : tonality for tonality in model.TONALITIES.keys() }, translate = False))
descr.def_attr("instrument_tonality", field.EnumField({ _("tonality_%s" % tonality) : tonality for tonality in model.TONALITIES.keys() }, translate = False))
descr.def_attr("view"      , field.HiddenField)
descr.def_attr("song"      , field.HiddenField)
descr.def_attr("notes"     , field.HiddenField)
descr.def_attr("view_type" , field.EnumField(set(VIEW_TYPE_NAMES.keys()), value_2_enum = view_value_2_enum, enum_2_value = view_enum_2_value))
descr.def_attr("capo"      , field.IntField)
descr.def_attr("g_key"     , field.BoolField)
descr.def_attr("f_key"     , field.BoolField)
descr.def_attr("g8"        , field.BoolField)
descr.def_attr("print_with_staff_too", field.BoolField)
descr.def_attr("volume"    , field.RangeField(0, 255))
descr.def_attr("muted"     , field.BoolField)
descr.def_attr("chorus"    , field.RangeField(0, 127))
descr.def_attr("reverb"    , field.RangeField(0, 127))
descr.def_attr("let_ring"  , field.BoolField)
descr.def_attr("visible"   , field.HiddenField, get_method = lambda partition: partition.view.visible, set_method = lambda partition, visible: setattr(partition.view, "visible", visible))
descr.def_attr("tuning"    , field.ObjectSelectorField, get_method = get_tuning)


def string_2_icon_filename(self):
  return os.path.join(globdef.DATADIR, "string_%s.png" % self.width())
  
descr = introsp.description(model.TablatureString)
descr.set_icon_filename(string_2_icon_filename)
descr.def_attr("notation_pos", field.HiddenField)
descr.def_attr("base_note"   , field.HiddenField)
descr.def_attr("octavo"      , field.EnumField(list(range(10))))
descr.def_attr("base_value"  , field.EnumField(note_2_base_note_names, translate = False))
descr.set_constructor(introsp.Constructor(new_tab_string))

#descr = introsp.description(model.DrumString)
#descr.set_icon_filename(os.path.join(globdef.DATADIR, "tamtam.png"))
#descr.def_attr("notation_pos", field.HiddenField)


def note_2_icon_filename(self):
  filename = os.path.join(globdef.DATADIR, "note_%s.png" % self.duration)
  if os.path.exists(filename): return filename
  return os.path.join(globdef.DATADIR, "note_96.png")


descr = introsp.description(model.Note)
descr.set_icon_filename(note_2_icon_filename)
descr.def_attr("string_id"   , field.HiddenField)
descr.def_attr("time"        , field.HiddenField)
descr.def_attr("partition"   , field.HiddenField)
descr.def_attr("button_rank" , field.HiddenField)
descr.def_attr("bellows_dir" , field.HiddenField)
descr.def_attr("duration"    , DurationField)
descr.def_attr("value"       , field.HiddenField)
descr.def_attr("base_value"  , field.EnumField(note_2_base_note_names, translate = False))
descr.def_attr("octavo"      , field.EnumField(list(range(10))))
descr.def_attr("fx"          , field.EnumField(dict((_(fx or "(none)"), fx) for fx in model.FXS          )), optional = False)
descr.def_attr("link_fx"     , field.EnumField(dict((_(fx or "(none)"), fx) for fx in model.LINK_FXS     )), optional = False)
descr.def_attr("duration_fx" , field.EnumField(dict((_(fx or "(none)"), fx) for fx in model.DURATION_FXS )), optional = False)
descr.def_attr("strum_dir_fx", field.EnumField(dict((_(fx or "(none)"), fx) for fx in model.STRUM_DIR_FXS)), optional = False)
descr.def_attr("volume"      , field.RangeField(0, 255))

descr = introsp.description(model.Mesure)
descr.set_icon_filename(os.path.join(globdef.DATADIR, "bar.png"))
descr.def_attr("number"   , field.HiddenField)
descr.def_attr("time"     , field.HiddenField)
descr.def_attr("duration" , field.HiddenField)
descr.def_attr("song"     , field.HiddenField)
descr.def_attr("rythm1"   , field.EnumField([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]))
descr.def_attr("rythm2"   , field.EnumField([4, 8]))
descr.def_attr("tempo"    , field.IntField)
descr.def_attr("syncope"  , field.BoolField)

def new_playlist_item(Class, playlist):
  import songwrite3.main as main
  app = main.APPS[playlist.song]
  mesure = app.selected_mesure
  if isinstance(mesure, model.Mesure): start = end = mesure.get_number()
  else:
    mesure_numbers = [mesure.get_number() for mesure in mesure.objects]
    start = min(mesure_numbers)
    end   = max(mesure_numbers)
  return Class(playlist, start, end)

descr = introsp.description(model.Playlist)
descr.set_icon_filename(os.path.join(globdef.DATADIR, "playlist.png"))
descr.def_attr("playlist_items", addable_values = [introsp.NewInstanceOf(model.PlaylistItem)], add_method = "insert", remove_method = "remove", label = "")
descr.def_attr("symbols"       , field.HiddenField)
descr.def_attr("song"          , field.HiddenField)

descr = introsp.description(model.PlaylistItem)
descr.set_icon_filename(os.path.join(globdef.DATADIR, "playlist.png"))
descr.def_attr("playlist"    , field.HiddenField)
descr.def_attr("from_mesure" , field.HiddenField)
descr.def_attr("to_mesure"   , field.HiddenField)
descr.def_attr("from_mesure1", field.IntField)
descr.def_attr("to_mesure1"  , field.IntField)
descr.set_constructor(introsp.Constructor(new_playlist_item))

descr = introsp.description(model.Lyrics)
descr.set_icon_filename(os.path.join(globdef.DATADIR, "lyrics.png"))
descr.def_attr("text"  , field.HiddenField)
descr.def_attr("melody", field.HiddenField)
descr.def_attr("song"  , field.HiddenField)
descr.def_attr("header", field.TextField)
descr.def_attr("show_all_lines_on_melody"  , field.BoolField)
descr.def_attr("show_all_lines_after_score", field.BoolField)



# -*- coding: utf-8 -*-

# Songwrite 3
# Copyright (C) 2007-2016 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, os.path

import songwrite3.globdef as globdef
import songwrite3.model   as model

PLUGINS = {}
class Plugin(object):
  def __init__(self, name):
    print("Registering %s plugin..." % name, file = sys.stderr)
    
    self.name = name
    
    PLUGINS[name] = self
    
  def create_ui(self, app):
    pass


EXPORT_FORMATS = []
class ExportPlugin(Plugin):
  def __init__(self, format, exts, can_export_song = 1, can_export_songbook = 0):
    self.format              = format
    self.exts                = exts
    self.can_export_song     = can_export_song
    self.can_export_songbook = can_export_songbook
    EXPORT_FORMATS.append(format)
    Plugin.__init__(self, "%s exporter" % format.lower())
    
  def create_ui(self, app):
    if self.can_export_song:
      app.add_to_menu(app.export_menu, 0, _("Export to %s") % self.format + "..." * int(globdef.config.ASK_FILENAME_ON_EXPORT), self.on_menu_click, arg = app)
    if self.can_export_songbook:
      app.add_to_menu(app.export_songbook_menu, 0, _("Export songbook to %s") % self.format + "..." * int(globdef.config.ASK_FILENAME_ON_EXPORT), self.on_songbook_menu_click, arg = app)
      
  def on_menu_click(self, widget, app):
    if globdef.config.ASK_FILENAME_ON_EXPORT or (not app.filename):
      filename = app.prompt_save_filename(self.exts, self.format)
    else:
      filename = app.filename.replace(".sw.xml", "") + "." + self.exts[0]
    if filename: self.export_to(app.song, filename)
    
  def on_songbook_menu_click(self, widget, app):
    if globdef.config.ASK_FILENAME_ON_EXPORT or (not app.songbook.filename):
      filename = app.prompt_save_filename(self.exts, self.format)
    else:
      filename = app.songbook.filename.replace(".sw.xml", "") + "." + self.exts[0]
    if filename: self.export_to(app.songbook, filename)
    
  def export_to(self, song, filename):
    data = self.export_to_string(song)
    if isinstance(data, bytes):
      if filename == "-": sys.stdout.buffer.write(data)
      else: open(filename, "wb").write(data)
    else:
      if filename == "-": print(data)
      else: open(filename, "w").write(data)
      
  def export_to_string(self, song): pass
  
  
IMPORT_FORMATS = []
class ImportPlugin(Plugin):
  def __init__(self, format, exts, binary = False):
    self.format = format
    self.exts   = exts
    self.binary = binary
    IMPORT_FORMATS.append(format)
    Plugin.__init__(self, "%s importer" % format.lower())
    
  def create_ui(self, app):
    app.add_to_menu(app.import_menu, 0, _("Import from %s") % self.format + "...", self.on_menu_click, arg = app)
    
  def on_menu_click(self, widget, app):
    if app.check_save(): return
    
    filename = app.prompt_open_filename(self.exts, self.format)
    
    if filename:
      song = self.import_from(filename)
      app.set_song(song)
      
  def import_from(self, filename):
    if filename == "-":
      return self.import_from_string(sys.stdin.read())
    else:
      if self.binary: s = open(filename, "rb").read()
      else:           s = open(filename      ).read()
    return self.import_from_string(s)
    
  def import_from_string(self, data): pass


PLUGINS_VIEWS = {}
class ViewPlugin(Plugin):
  def __init__(self, View, String, code_name, category):
    PLUGINS_VIEWS[code_name] = View, String
    Plugin.__init__(self, "%s view" % code_name)
    
    if category: model.VIEWS[category].append(View)
    
    
def create_plugins_ui(app):
  for plugin in PLUGINS.values(): plugin.create_ui(app)
  
def load_all_plugins():
  plugins_dir = os.path.dirname(__file__)
  for file in [
    "songwrite",
    "abc",
    "texttab",
    "midi",
    "pdf",
    "fingering",
    "additional_instruments",
    "lyre",
    "chord",
    "accordion",
    ]:
    try:
      __import__("songwrite3.plugins.%s" % file)
    except:
      print("Error while loading plugin 'songwrite3.plugins.%s'!" % file, file = sys.stderr)
      sys.excepthook(*sys.exc_info())
  print(file = sys.stderr)
  
def get_importer(format):
  return PLUGINS["%s importer" % format.lower()]

def get_exporter(format):
  return PLUGINS["%s exporter" % format.lower()]

def load_view(partition, attrs):
  View, String = PLUGINS_VIEWS[attrs["type"]]
  view = View(partition, "")
  view.load_attrs(attrs)
  return view, String
  
